# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


from collections import deque
from itertools import combinations

import networkx as nx
import pandas as pd

from pandapower.topology.create_graph import create_nxgraph


def connected_component(mg, bus, notravbuses=[]):
    """
    Finds all buses in a NetworkX graph that are connected to a certain bus.

    INPUT:
        **mg** (NetworkX graph) - NetworkX Graph or MultiGraph that represents a pandapower network.

        **bus** (integer) - Index of the bus at which the search for connected components originates


    OPTIONAL:
     **notravbuses** (list/set) - indices of notravbuses: lines connected to these buses are
                                     not being considered in the graph

    OUTPUT:
        **cc** (generator) - Returns a generator that yields all buses connected to the input bus

    EXAMPLE:
         >>> from pandapower.topology.create_graph import create_nxgraph
         >>> from pandapower.topology.graph_searches import connected_component
         >>> mg = create_nxgraph(net)
         >>> cc = connected_component(mg, 5)
    """
    yield bus
    visited = {bus}
    stack = deque([iter(mg[bus])])
    while stack:
        for child in stack.pop():
            if child not in visited:
                yield child
                visited.add(child)
                if child not in notravbuses:
                    stack.append(iter(mg[child]))


def connected_components(mg, notravbuses=set()):
    """
     Clusters all buses in a NetworkX graph that are connected to each other.

     INPUT:
        **mg** (NetworkX graph) - NetworkX Graph or MultiGraph that represents a pandapower network.


     OPTIONAL:
     **notravbuses** (set) - Indices of notravbuses: lines connected to these buses are
     not being considered in the graph

     OUTPUT:
        **cc** (generator) - Returns a generator that yields all clusters of buses connected
                             to each other.

     EXAMPLE:
         >>> from pandapower.topology.create_graph import create_nxgraph
         >>> from pandapower.topology.graph_searches import connected_components
         >>> mg = create_nxgraph(net)
         >>> cc = connected_components(mg, 5)
    """

    nodes = set(mg.nodes()) - notravbuses
    while nodes:
        cc = set(connected_component(mg, nodes.pop(), notravbuses=notravbuses))
        yield cc
        nodes -= cc
    # the above does not work if two notravbuses are directly connected
    if len(notravbuses) > 0:
        for f, t in mg.edges(notravbuses):
            if f in notravbuses and t in notravbuses:
                yield set([f, t])


def calc_distance_to_bus(net, bus, respect_switches=True, nogobuses=None,
                         notravbuses=None, weight='weight', g=None):
    """
        Calculates the shortest distance between a source bus and all buses connected to it.

     INPUT:
        **net** (pandapowerNet) - Variable that contains a pandapower network.

        **bus** (integer) - Index of the source bus.


     OPTIONAL:
        **respect_switches** (boolean, True)

            True: open line switches are being considered (no edge between nodes).

            False: open line switches are being ignored.

        **nogobuses** (integer/list, None) - nogobuses are not being considered.

        **notravbuses** (integer/list, None) - lines connected to these buses are not being considered.

        **weight** (string, None) – Edge data key corresponding to the edge weight.

        **g** (nx.MultiGraph, None) – MultiGraph of the network. If None, the graph will be created.

     OUTPUT:
        **dist** - Returns a pandas series with containing all distances to the source bus
                   in km. If weight=None dist is the topological distance (int).

     EXAMPLE:
         >>> from pandapower.topology.graph_searches import calc_distance_to_bus
         >>> dist = calc_distance_to_bus(net, 5)

    """
    if g is None:
        g = create_nxgraph(net, respect_switches=respect_switches, nogobuses=nogobuses,
                           notravbuses=notravbuses)
    return pd.Series(nx.single_source_dijkstra_path_length(g, bus, weight=weight))


def unsupplied_buses(net, mg=None, slacks=None, respect_switches=True):
    """
     Finds buses, that are not connected electrically (no lines, trafos etc or if respect_switches
     is True only connected via open switches) to an external grid and that are in service.

     INPUT:
        **net** (pandapowerNet) - variable that contains a pandapower network

     OPTIONAL:
        **mg** (NetworkX graph) - NetworkX Graph or MultiGraph that represents a pandapower network.

        **in_service_only** (boolean, False) - Defines whether only in service buses should be
            included in unsupplied_buses.

        **slacks** (set, None) - buses which are considered as root / slack buses. If None, all
            existing slack buses are considered.

        **respect_switches** (boolean, True) - Fixes how to consider switches - only in case of no
            given mg.

     OUTPUT:
        **ub** (set) - unsupplied buses

     EXAMPLE:
         >>> from pandapower.topology.graph_searches import unsupplied_buses
         >>> unsupplied_buses(net)
    """

    mg = mg or create_nxgraph(net, respect_switches=respect_switches)
    if slacks is None:
        slacks = (set(net.ext_grid[net.ext_grid.in_service].bus.values) |
                  set(net.gen[net.gen.in_service & net.gen.slack].bus.values) |
                  set(net.vsc[net.vsc.in_service & (net.vsc.control_mode_ac == "slack")].bus.values))
    not_supplied = set()
    for cc in nx.connected_components(mg):
        if not set(cc) & slacks:
            not_supplied.update(set(cc))

    return not_supplied


def find_basic_graph_characteristics(g, roots, characteristics):
    """
    Determines basic characteristics of the given graph like connected buses, stubs, bridges,
    and articulation points.

    .. note::
        This is the base function for find_graph_characteristics. Please use the latter
        function instead!
    """
    connected = 'connected' in characteristics
    stub_buses = 'stub_buses' in characteristics
    bridges = {'bridges', 'required_bridges'} & set(characteristics)
    articulation_points = {'articulation_points', 'notn1_areas'} & set(characteristics)
    notn1_starts = 'notn1_areas' in characteristics

    char_dict = {'connected': set(), 'stub_buses': set(), 'bridges': set(),
                 'articulation_points': set(), 'notn1_starts': set()}

    discovery = {root: 0 for root in roots}  # "time" of first discovery of node during search
    low = {root: 0 for root in roots}
    visited = set(roots)
    path = []
    stack = [(root, root, iter(g[root])) for root in roots]
    while stack:
        grandparent, parent, children = stack[-1]
        try:
            child = next(children)
            if stub_buses:
                if child not in visited:
                    path.append(child)  # keep track of movement through the graph
            if grandparent == child:
                continue
            if child in visited:
                if discovery[child] <= discovery[parent]:  # back edge
                    low[parent] = min(low[parent], discovery[child])
            else:
                low[child] = discovery[child] = len(discovery)
                visited.add(child)
                stack.append((parent, child, iter(g[child])))
        except StopIteration:
            back = stack.pop()
            path.append(back[0])
            if low[parent] >= discovery[grandparent]:
                # Articulation points and start of not n-1 safe buses
                if grandparent not in roots:
                    if articulation_points:
                        char_dict['articulation_points'].add(grandparent)
                    if notn1_starts:
                        char_dict['notn1_starts'].add(parent)
                if low[parent] > discovery[grandparent]:
                    # Bridges
                    if bridges:
                        char_dict['bridges'].add((grandparent, parent))

                    # Stub buses
                    if stub_buses:
                        stub = path.pop()
                        if stub != grandparent:
                            char_dict['stub_buses'].add(stub)
                        while path and path[-1] != grandparent and path[-1] not in roots:
                            stub = path.pop()
                            char_dict['stub_buses'].add(stub)
            low[grandparent] = min(low[parent], low[grandparent])

    if connected:
        char_dict['connected'] = visited
    return char_dict


def find_graph_characteristics(g, roots, characteristics):
    """
    Finds and returns different characteristics of the given graph which can be specified.

    INPUT:
        **g** (NetworkX graph) - Graph of the network

        **roots** (list) - Root buses of the graphsearch

        **characteristics** (list) - List of characteristics this function determines and returns

        .. note::

            Possible characteristics:

            - 'connected' - All buses which have a connection to at least one of the root buses
            - 'articulation_points' - Buses which lead to disconnected areas if they get removed
            - 'bridges' - Edges which lead to disconnected areas if they get removed
            - 'stub_buses' - Buses which arent't connected if one specific edge gets removed
            - 'required_bridges' - Bridges which are strictly needed to connect a specific bus
            - 'notn1_areas' - Areas which aren't connected if one specific bus gets removed

    OUTPUT:

        **char_dict** (dict) - dictionary which contains the wanted characteristics

        ======================= ================================================================
        key                     dict value
        ======================= ================================================================
        'connected'             set of all connected buses
        'articulation_points'   set of all articulation points
        'bridges'               set of tuples which represent start and end bus of each bridge
        'stub_buses'            set of all buses which lie on a stub
        'required_bridges'      dict of all buses which are connected via at least one bridge.
                                The dict values contain a set of bridges which are needed to
                                connect the key buses
        'notn1_areas'           dict of not n-1 safe areas. The dict values contain a set of
                                not n-1 safe buses which aren't connected if the key bus gets
                                removed
        ======================= ================================================================

    EXAMPLE::

        import topology as top
        g = top.create_nxgraph(net, respect_switches=False)
        char_dict = top.find_graph_characteristics(g, roots=[0, 3], characteristics=['connected', 'stub_buses'])
    """
    char_dict = find_basic_graph_characteristics(g, roots, characteristics)

    required_bridges = 'required_bridges' in characteristics
    notn1_areas = 'notn1_areas' in characteristics

    if not required_bridges and not notn1_areas:
        return {key: char_dict[key] for key in characteristics}

    char_dict.update({'required_bridges': dict(), 'notn1_areas': dict()})

    visited = set(roots)
    visited_bridges = []
    notn1_area_start = None
    curr_notn1_area = []

    stack = [(root, root, iter(g[root])) for root in roots]
    while stack:
        grandparent, parent, children = stack[-1]
        try:
            child = next(children)
            if child == grandparent:
                continue
            if child not in visited:
                visited.add(child)
                stack.append((parent, child, iter(g[child])))
                if required_bridges and ((parent, child) in char_dict['bridges'] or
                                         (child, parent) in char_dict['bridges']):
                    visited_bridges.append((parent, child))

                if notn1_areas:
                    if child in char_dict['notn1_starts'] and not notn1_area_start:
                        notn1_area_start = parent
                    if notn1_area_start:
                        curr_notn1_area.append(child)

        except StopIteration:
            stack.pop()
            if required_bridges:
                if len(visited_bridges) > 0:
                    char_dict['required_bridges'][parent] = visited_bridges[:]
                if ((parent, grandparent) in char_dict['bridges'] or
                        (grandparent, parent) in char_dict['bridges']):
                    visited_bridges.pop()

            if notn1_areas and grandparent == notn1_area_start:
                if grandparent in char_dict["notn1_areas"]:
                    char_dict["notn1_areas"][grandparent].update(set(curr_notn1_area[:]))
                else:
                    char_dict["notn1_areas"][grandparent] = set(curr_notn1_area[:])
                del curr_notn1_area[:]
                notn1_area_start = None

    return {key: char_dict[key] for key in characteristics}


def get_2connected_buses(g, roots):
    """
    Get all buses which have at least two connections to the roots

    INPUT:
        **g** (NetworkX graph) - NetworkX Graph or MultiGraph that represents a pandapower network

        **roots** - Roots of the graphsearch
    """
    char_dict = find_graph_characteristics(g, roots, characteristics=['connected', 'stub_buses'])
    connected, stub_buses = char_dict['connected'], char_dict['stub_buses']
    two_connected = connected - stub_buses
    return connected, two_connected


def determine_stubs(net, roots=None, mg=None, respect_switches=False):
    """
     Finds stubs in a network. Open switches are being ignored. Results are being written in a new
     column in the bus table ("on_stub") and line table ("is_stub") as True/False value.


     INPUT:
        **net** (pandapowerNet) - Variable that contains a pandapower network.

     OPTIONAL:
        **roots** (integer/list, None) - indices of buses that should be excluded (by default, the
                                         ext_grid buses will be set as roots)

     EXAMPLE:
         >>> from pandapower.topology.graph_searches import determine_stubs
         >>> determine_stubs(net, roots = [0, 1])
    """
    if mg is None:
        mg = create_nxgraph(net, respect_switches=respect_switches)
    # remove buses with degree lower 2 until none left
    if roots is None:
        roots = set(net.ext_grid.bus)
    #    mg.add_edges_from((a, b) for a, b in zip(list(roots)[:-1], list(roots)[1:]))
    #    while True:
    #        dgo = {g for g, d in list(mg.degree().items()) if d < 2} #- roots
    #        if not dgo:
    #            break
    #        mg.remove_nodes_from(dgo)
    #    n1_buses = mg.nodes()
    _, n1_buses = get_2connected_buses(mg, roots)
    net.bus["on_stub"] = True
    net.bus.loc[list(n1_buses), "on_stub"] = False
    net.line["is_stub"] = ~((net.line.from_bus.isin(n1_buses)) & (net.line.to_bus.isin(n1_buses)))
    stubs = set(net.bus.index) - set(n1_buses)
    return stubs


def lines_on_path(mg, path):
    """
     Finds all lines that connect a given path of buses.

     INPUT:
        **mg** (NetworkX graph) - NetworkX Graph or MultiGraph that represents a pandapower network.

        **path** (list) - List of connected buses.

     OUTPUT:
        **lines** (list) - Returns a list of all lines on the path.

     EXAMPLE:
         import topology as top

         mg = top.create_nxgraph(net)
         lines = top.lines_on_path(mg, [4, 5, 6])

     """

    return elements_on_path(mg, path, "line")


def elements_on_path(mg, path, element="line"):
    """
     Finds all elements that connect a given path of buses.

     INPUT:
        **mg** (NetworkX graph) - NetworkX Graph or MultiGraph that represents a pandapower network.

        **path** (list) - List of connected buses.

        **element** (string, "l") - element type

        **multi** (boolean, True) - True: Applied on a NetworkX MultiGraph
                                    False: Applied on a NetworkX Graph

     OUTPUT:
        **elements** (list) - Returns a list of all lines on the path.

     EXAMPLE:
         import topology as top

         mg = top.create_nxgraph(net)
         elements = top.elements_on_path(mg, [4, 5, 6])

     """
    if element not in ["line", "switch", "trafo", "trafo3w"]:
        raise ValueError("Invalid element type %s" % element)
    if isinstance(mg, nx.MultiGraph):
        return [edge[1] for b1, b2 in zip(path, path[1:]) for edge in mg.get_edge_data(b1, b2).keys()
                if edge[0] == element]
    else:
        return [mg.get_edge_data(b1, b2)["key"][1] for b1, b2 in zip(path, path[1:])
                if mg.get_edge_data(b1, b2)["key"][0] == element]


def get_end_points_of_continuously_connected_lines(net, lines):
    mg = nx.MultiGraph()
    line_buses = net.line.loc[lines, ["from_bus", "to_bus"]].values
    mg.add_edges_from(line_buses)
    switch_buses = net.switch[["bus", "element"]].values[net.switch.et.values == "b"]
    mg.add_edges_from(switch_buses)

    all_buses = set(line_buses.flatten())
    longest_path = []
    for b1, b2 in combinations(all_buses, 2):
        try:
            path = nx.shortest_path(mg, b1, b2)
        except nx.NetworkXNoPath:
            raise UserWarning("Lines not continuously connected")
        if len(path) > len(longest_path):
            longest_path = path
    if all_buses - set(longest_path):
        raise UserWarning("Lines have branching points")
    return longest_path[0], longest_path[-1]



# -*- coding: utf-8 -*-

# Copyright 1996-2015 PSERC. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


import numpy as np
from numba import jit
from scipy.sparse import csr_matrix, coo_matrix

from pandapower.pypower.idx_brch import F_BUS, T_BUS
from pandapower.pypower.idx_bus import GS, BS
from pandapower.pypower.makeYbus import branch_vectors


@jit(nopython=True, cache=False)
def gen_Ybus(Yf_x, Yt_x, Ysh, col_Y, f, t, f_sort, t_sort, nb, nl, r_nl):  # pragma: no cover
    """
    Fast calculation of Ybus
    """

    r_nb = range(nb)

    # allocate data of Ybus in CSR format
    # Note: More space is allocated than needed with empty.
    #       The matrix size will be reduced afterwards
    alloc_size = nl * 2 + nb
    Yx = np.empty(alloc_size, dtype=np.complex128)  # data
    Yp = np.zeros(nb + 1, dtype=np.int64)  # row pointer
    Yj = np.empty(alloc_size, dtype=np.int64)  # colum indices

    # index iterators
    # a = iterator of f, b = iterator of t, curRow = current Row
    a, b, curRow = 0, 0, 0
    # number of nonzeros (total), number of nonzeros per row
    nnz, nnz_row = 0, 0
    # flag checks if diagonal entry was added
    YshAdded = False

    for curRow in r_nb:
        nnz_row = 0
        # iterate rows of Ybus

        # add entries from Yf
        while a < nl and f[f_sort[a]] == curRow:
            # Entries from f_sort[a] in current row of Ybus
            for col in (r_nl[f_sort[a]], r_nl[f_sort[a]] + nl):
                # 'Has entry at column in Yf: %i ' % col
                if col_Y[col] == curRow and not YshAdded:
                    # add Ysh and Yf_x (diagonal element). If not already added
                    curVal = Yf_x[col] + Ysh[curRow]
                    YshAdded = True
                else:
                    # add only Yf_x
                    curVal = Yf_x[col]

                for k in range(Yp[curRow], Yp[curRow] + nnz_row):
                    if col_Y[col] == Yj[k]:
                        # if entry at column already exists add value
                        Yx[k] += curVal
                        break
                else:
                    # new entry in Ybus
                    Yx[nnz] = curVal
                    Yj[nnz] = col_Y[col]
                    nnz += 1
                    nnz_row += 1
            a += 1

        # add entries from Yt
        while b < nl and t[t_sort[b]] == curRow:
            # Entries from t_sort[b] in current row of Ybus
            for col in (r_nl[t_sort[b]], r_nl[t_sort[b]] + nl):
                # 'Has entry at column in Yt: %i ' % col
                if col_Y[col] == curRow and not YshAdded:
                    # add Ysh and Yf_x (diagonal element). If not already added
                    curVal = Yt_x[col] + Ysh[curRow]
                    YshAdded = True
                else:
                    # add only Yt_x
                    curVal = Yt_x[col]

                for k in range(Yp[curRow], Yp[curRow] + nnz_row):
                    if col_Y[col] == Yj[k]:
                        # if entry at column already exists add value
                        Yx[k] += curVal
                        break
                else:
                    # new entry in Ybus
                    Yx[nnz] = curVal
                    Yj[nnz] = col_Y[col]
                    nnz += 1
                    nnz_row += 1
            b += 1

        if not YshAdded:
            # check if diagonal entry was added. If not -> add if not zero
            if Ysh[curRow]:
                Yx[nnz] = Ysh[curRow]
                Yj[nnz] = curRow
                nnz += 1
                nnz_row += 1

        YshAdded = False
        # add number of nonzeros in row to row pointer
        Yp[curRow + 1] = nnz_row + Yp[curRow]
        curRow += 1

    return Yx, Yj, Yp, nnz


def makeYbus(baseMVA, bus, branch):
    """Builds the bus admittance matrix and branch admittance matrices.

    Returns the full bus admittance matrix (i.e. for all buses) and the
    matrices C{Yf} and C{Yt} which, when multiplied by a complex voltage
    vector, yield the vector currents injected into each line from the
    "from" and "to" buses respectively of each line. Does appropriate
    conversions to p.u.

    @see: L{makeSbus}

    @author: Ray Zimmerman (PSERC Cornell)
    @author: Richard Lincoln

    modified by Florian Schaefer (to use numba) (florian.schaefer@uni-kassel.de)
    """
    ## constants
    nb = bus.shape[0]  ## number of buses
    nl = branch.shape[0]  ## number of lines

    ## for each branch, compute the elements of the branch admittance matrix where
    ##
    ##      | If |   | Yff  Yft |   | Vf |
    ##      |    | = |          | * |    |
    ##      | It |   | Ytf  Ytt |   | Vt |
    ##
    Ytt, Yff, Yft, Ytf = branch_vectors(branch, nl)

    ## compute shunt admittance
    ## if Psh is the real power consumed by the shunt at V = 1.0 p.u.
    ## and Qsh is the reactive power injected by the shunt at V = 1.0 p.u.
    ## then Psh - j Qsh = V * conj(Ysh * V) = conj(Ysh) = Gs - j Bs,
    ## i.e. Ysh = Psh + j Qsh, so ...
    ## vector of shunt admittances
    Ysh = (bus[:, GS] + 1j * bus[:, BS]) / baseMVA

    ## build connection matrices
    f = np.real(branch[:, F_BUS]).astype(np.int64)  ## list of "from" buses
    t = np.real(branch[:, T_BUS]).astype(np.int64)  ## list of "to" buses

    ## build Yf and Yt such that Yf * V is the vector of complex branch currents injected
    ## at each branch's "from" bus, and Yt is the same for the "to" bus end
    i = np.hstack([np.arange(nl), np.arange(nl)])  ## double set of row indices

    Yf_x = np.hstack([Yff, Yft])
    Yt_x = np.hstack([Ytf, Ytt])
    col_Y = np.hstack([f, t])

    Yf = coo_matrix((Yf_x, (i, col_Y)), (nl, nb)).tocsr()
    Yt = coo_matrix((Yt_x, (i, col_Y)), (nl, nb)).tocsr()
    Yx, Yj, Yp, nnz = gen_Ybus(Yf_x, Yt_x, Ysh, col_Y, f, t, np.argsort(f), np.argsort(t), nb, nl,
                               np.arange(nl, dtype=np.int64))
    Ybus = csr_matrix((np.resize(Yx, nnz), np.resize(Yj, nnz), Yp), (nb, nb))
    for Y in (Ybus, Yf, Yt):
        Y.sort_indices()
        Y.eliminate_zeros()
    return Ybus, Yf, Yt



import os
pp_dir = os.path.dirname(os.path.realpath(__file__))

from pandapower._version import __version__, __format_version__
from pandapower.auxiliary import *  # TODO: some functions shouldn't be available with import pandapower
from pandapower.std_types import *
from pandapower.create import *
from pandapower.convert_format import *
from pandapower.file_io import *
from pandapower.sql_io import to_postgresql, from_postgresql, delete_postgresql_net, to_sqlite, from_sqlite
from pandapower.powerflow import *
from pandapower.optimal_powerflow import OPFNotConverged
from pandapower.run import *
from pandapower.toolbox import *  # to be removed -> in future via package namespace available
from pandapower.groups import *
from pandapower.diagnostic import *
from pandapower.runpm import *
from pandapower.pf.runpp_3ph import runpp_3ph

import geojson
geojson.geometry.DEFAULT_PRECISION = 8

import pandas as pd
pd.options.mode.chained_assignment = None  # default='warn'

# import pandapower packages
import pandapower.control
import pandapower.converter
import pandapower.estimation
import pandapower.grid_equivalents
import pandapower.networks
# import pandapower.opf  # no imports available yet
# import pandapower.pf  # no imports available yet
import pandapower.plotting
# import pandapower.protection  # no imports available yet
# import pandapower.pypower  # no imports available yet
import pandapower.shortcircuit
import pandapower.timeseries
import pandapower.toolbox
import pandapower.topology



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


from time import perf_counter

from numpy import flatnonzero as find, r_, zeros, argmax, setdiff1d, union1d, any, int32, \
    sum as np_sum, abs as np_abs, int64

from pandapower.pf.ppci_variables import _get_pf_variables_from_ppci, _store_results_from_pf_in_ppci
from pandapower.pf.run_dc_pf import _run_dc_pf
from pandapower.pypower.bustypes import bustypes
from pandapower.pypower.idx_bus import BUS_I, PD, QD, BUS_TYPE, PQ, PV, GS, BS, SL_FAC as SL_FAC_BUS
from pandapower.pypower.idx_gen import PG, QG, QMAX, QMIN, GEN_BUS, GEN_STATUS, SL_FAC
from pandapower.pypower.makeSbus import makeSbus
from pandapower.pypower.makeYbus import makeYbus as makeYbus_pypower
from pandapower.pypower.newtonpf import newtonpf
from pandapower.pypower.pfsoln import _update_v
from pandapower.pypower.pfsoln import pfsoln as pfsoln_pypower
from pandapower.auxiliary import version_check

try:
    from pandapower.pf.makeYbus_numba import makeYbus as makeYbus_numba
    from pandapower.pf.pfsoln_numba import pfsoln as pfsoln_numba, pf_solution_single_slack
    version_check('numba')
    numba_installed = True
except ImportError:
    numba_installed = False

try:
    from lightsim2grid.newtonpf import newtonpf_new as newton_ls
except ImportError:
    newton_ls = None


def _run_newton_raphson_pf(ppci, options):
    """
    Runs a Newton-Raphson power flow.

    INPUT
    ppci (dict) - the "internal" ppc (without out ot service elements and sorted elements)
    options(dict) - options for the power flow

    """
    t0 = perf_counter()
    # we cannot run DC pf before running newton with distributed slack because the slacks come pre-solved after the DC pf
    if isinstance(options["init_va_degree"], str) and options["init_va_degree"] == "dc":
        if options['distributed_slack']:
            pg_copy = ppci['gen'][:, PG].copy()
            pd_copy = ppci['bus'][:, PD].copy()
            ppci = _run_dc_pf(ppci, options["recycle"])
            ppci['gen'][:, PG] = pg_copy
            ppci['bus'][:, PD] = pd_copy
        else:
            ppci = _run_dc_pf(ppci, options["recycle"])
    if options["enforce_q_lims"]:
        ppci, success, iterations, bus, gen, branch = _run_ac_pf_with_qlims_enforced(ppci, options)
    else:
        ppci, success, iterations = _run_ac_pf_without_qlims_enforced(ppci, options)
        # update data matrices with solution store in ppci
        bus, gen, branch = ppci_to_pfsoln(ppci, options)
    et = perf_counter() - t0
    ppci = _store_results_from_pf_in_ppci(ppci, bus, gen, branch, success, iterations, et)
    return ppci


def ppci_to_pfsoln(ppci, options, limited_gens=None):
    internal = ppci["internal"]
    if options["only_v_results"]:
        # time series relevant hack which ONLY saves V from ppci
        _update_v(internal["bus"], internal["V"])
        return internal["bus"], internal["gen"], internal["branch"]
    else:
        # reads values from internal ppci storage to bus, gen, branch and returns it
        if options['distributed_slack']:
            # consider buses with non-zero slack weights as if they were slack buses,
            # and gens with non-zero slack weights as if they were reference machines
            # this way, the function pfsoln will extract results for distributed slack gens, too
            # also, the function pfsoln will extract results for the PQ buses for xwards
            gens_with_slack_weights = find(internal["gen"][:, SL_FAC] != 0)
            # gen_buses_with_slack_weights = internal["gen"][gens_with_slack_weights, GEN_BUS].astype(int32)
            buses_with_slack_weights = internal["bus"][find(internal["bus"][:, SL_FAC_BUS] != 0), BUS_I].astype(int32)
            # buses_with_slack_weights = union1d(gen_buses_with_slack_weights, buses_with_slack_weights)
            ref = union1d(internal["ref"], buses_with_slack_weights)
            ref_gens = union1d(internal["ref_gens"], gens_with_slack_weights)
        else:
            ref = internal["ref"]
            ref_gens = internal["ref_gens"]

        makeYbus, pfsoln = _get_numba_functions(ppci, options)

        # todo: this can be dropped if Ybus is returned from Newton and has the latest Ybus status:
        if options["tdpf"]:
            # needs to be updated to match the new R because of the temperature
            internal["Ybus"], internal["Yf"], internal["Yt"] = makeYbus(internal["baseMVA"], internal["bus"], internal["branch"])

        # todo: here Ybus_svc, Ybus_tcsc must be used (Ybus = Ybus + Ybus_svc + Ybus_tcsc)
        result_pfsoln = pfsoln(internal["baseMVA"], internal["bus"], internal["gen"], internal["branch"],
                               internal["svc"], internal["tcsc"], internal["ssc"], internal["vsc"],
                               internal["Ybus"], internal["Yf"], internal["Yt"], internal["V"],
                               ref, ref_gens, limited_gens=limited_gens)
        return result_pfsoln


def _get_Y_bus(ppci, options, makeYbus, baseMVA, bus, branch):
    recycle = options["recycle"]

    if isinstance(recycle, dict) and not recycle["trafo"] and ppci["internal"]["Ybus"].size:
        Ybus, Yf, Yt = ppci["internal"]['Ybus'], ppci["internal"]['Yf'], ppci["internal"]['Yt']
    else:
        # build admittance matrices
        Ybus, Yf, Yt = makeYbus(baseMVA, bus, branch)

    return ppci, Ybus, Yf, Yt


def _get_numba_functions(ppci, options):
    """
    pfsoln from pypower maybe slow in some cases. This function chooses the fastest for the given pf calculation
    """
    if options["numba"] and numba_installed:
        makeYbus = makeYbus_numba
        shunt_in_net = any(ppci["bus"][:, BS]) or any(ppci["bus"][:, GS])
        # faster pfsoln function if only one slack is in the grid and no gens
        pfsoln = pf_solution_single_slack if ppci["gen"].shape[0] == 1 \
                                             and not options["voltage_depend_loads"] \
                                             and not options['distributed_slack'] \
                                             and not shunt_in_net \
            else pfsoln_numba
    else:
        makeYbus = makeYbus_pypower
        pfsoln = pfsoln_pypower
    return makeYbus, pfsoln


def _store_internal(ppci, internal_storage):
    # internal storage is a dict with the variables to store in net["_ppc"]["internal"]
    for key, val in internal_storage.items():
        ppci["internal"][key] = val
    return ppci


def _get_Sbus(ppci, recycle=None):
    baseMVA, bus, gen = ppci["baseMVA"], ppci["bus"], ppci["gen"]
    if not isinstance(recycle, dict) or "Sbus" not in ppci["internal"]:
        return makeSbus(baseMVA, bus, gen)
    if recycle["bus_pq"] or recycle["gen"]:
        return makeSbus(baseMVA, bus, gen)
    return ppci["internal"]["Sbus"]


def _run_ac_pf_without_qlims_enforced(ppci, options):
    makeYbus, pfsoln = _get_numba_functions(ppci, options)

    baseMVA, bus, gen, branch, svc, tcsc, ssc, vsc, ref, pv, pq, *_, V0, ref_gens = _get_pf_variables_from_ppci(ppci, True)

    ppci, Ybus, Yf, Yt = _get_Y_bus(ppci, options, makeYbus, baseMVA, bus, branch)

    # compute complex bus power injections [generation - load]
    Sbus = _get_Sbus(ppci, options["recycle"])


    # run the newton power flow
    if options["lightsim2grid"]:
        V, success, iterations, J, Vm_it, Va_it = newton_ls(Ybus.tocsc(), Sbus, V0, ref, pv, pq, ppci, options)
        T = None
        r_theta_kelvin_per_mw = None
    else:
        V, success, iterations, J, Vm_it, Va_it, r_theta_kelvin_per_mw, T = newtonpf(Ybus, Sbus, V0, ref, pv, pq, ppci, options, makeYbus)
        # due to TPDF, SVC, TCSC, the Ybus matrices can be updated in the newtonpf and stored in ppci["internal"],
        # so we extract them here for later use:
        Ybus, Ybus_svc, Ybus_tcsc, Ybus_ssc, Ybus_vsc = (ppci["internal"].get(key) for key in ("Ybus", "Ybus_svc", "Ybus_tcsc", "Ybus_ssc", "Ybus_vsc"))
        Ybus = Ybus + Ybus_svc + Ybus_tcsc + Ybus_ssc + Ybus_vsc

    # keep "internal" variables in  memory / net["_ppc"]["internal"] -> needed for recycle.
    ppci = _store_internal(ppci, {"J": J, "Vm_it": Vm_it, "Va_it": Va_it, "bus": bus, "gen": gen, "branch": branch,
                                  "svc": svc, "tcsc": tcsc, "ssc": ssc, "vsc": vsc, "baseMVA": baseMVA, "V": V,
                                  "pv": pv, "pq": pq, "ref": ref,
                                  "Sbus": Sbus, "ref_gens": ref_gens, "Ybus": Ybus, "Yf": Yf, "Yt": Yt,
                                  "r_theta_kelvin_per_mw": r_theta_kelvin_per_mw, "T": T})

    return ppci, success, iterations


def _run_ac_pf_with_qlims_enforced(ppci, options):
    baseMVA, bus, gen, branch, svc, tcsc, ssc, vsc, ref, pv, pq, on, *_, V0, ref_gens = _get_pf_variables_from_ppci(ppci)
    bus_backup_p_q = bus[:, [PD, QD]].copy()
    gen_backup_p = gen[:, PG].copy()

    qlim = options["enforce_q_lims"]
    limited = []  # list of indices of gens @ Q lims
    fixedQg = zeros(gen.shape[0])  # Qg of gens at Q limits

    while True:
        ppci, success, iterations = _run_ac_pf_without_qlims_enforced(ppci, options)
        gen[:, PG] = gen_backup_p
        bus[:, PD] = bus_backup_p_q[:, 0]
        bus, gen, branch = ppci_to_pfsoln(ppci, options, limited)

        # find gens with violated Q constraints
        gen_status = gen[:, GEN_STATUS] > 0
        qg_max_lim = gen[:, QG] > gen[:, QMAX]
        qg_min_lim = gen[:, QG] < gen[:, QMIN]

        mx = setdiff1d(find(gen_status & qg_max_lim), ref_gens)
        mn = setdiff1d(find(gen_status & qg_min_lim), ref_gens)

        if len(mx) > 0 or len(mn) > 0:  # we have some Q limit violations
            # one at a time?
            if qlim == 2:  # fix largest violation, ignore the rest
                k = argmax(r_[gen[mx, QG] - gen[mx, QMAX],
                              gen[mn, QMIN] - gen[mn, QG]])
                if k > len(mx):
                    mn = mn[k - len(mx)]
                    mx = []
                else:
                    mx = mx[k]
                    mn = []

            # save corresponding limit values
            fixedQg[mx] = gen[mx, QMAX]
            fixedQg[mn] = gen[mn, QMIN]
            mx = r_[mx, mn].astype(int64)

            # convert to PQ bus
            gen[mx, QG] = fixedQg[mx]  # set Qg to binding
            #            if len(ref) > 1 and any(bus[gen[mx, GEN_BUS].astype(np.int64), BUS_TYPE] == REF):
            #                raise ValueError('Sorry, pandapower cannot enforce Q '
            #                                 'limits for slack buses in systems '
            #                                 'with multiple slacks.')

            changed_gens = gen[mx, GEN_BUS].astype(int64)
            bus[setdiff1d(changed_gens, ref), BUS_TYPE] = PQ  # & set bus type to PQ

            # update bus index lists of each type of bus
            ref, pv, pq = bustypes(bus, gen)

            limited = r_[limited, mx].astype(int64)

            for i in range(len(limited)):  # [one at a time, since they may be at same bus]
                gen[limited[i], GEN_STATUS] = 0  # temporarily turn off gen,
                bi = gen[limited[i], GEN_BUS].astype(int64)  # adjust load accordingly,
                bus[bi, [PD, QD]] = (bus[bi, [PD, QD]] - gen[limited[i], [PG, QG]])
        else:
            break  # no more generator Q limits violated

    if len(limited) > 0:
        # restore injections from limited gens [those at Q limits]
        bus[setdiff1d(changed_gens, ref), BUS_TYPE] = PV  # & set bus type back to PV
        gen[limited, QG] = fixedQg[limited]  # restore Qg value,
        gen[limited, GEN_STATUS] = 1  # turn gens back on
        bus[:, [PD, QD]] = bus_backup_p_q
    return ppci, success, iterations, bus, gen, branch



import numpy as np
import pandas as pd

def add_storage_opf_settings(net, ppci, pm):
    # callback function to add storage settings. Must be called after initializing pm data structure since the
    # pm["storage"] dict is filled

    # n time steps to optimize (here 3 hours)
    pm["n_time_steps"] = net._options["n_time_steps"]
    # time step (here 1 hour)
    pm["time_elapsed"] = net._options["time_elapsed"]

    # add storage systems to pm
    # Todo: Some variables are not used and not included in pandapower as well (energy_rating, thermal_rating,
    # (efficiencies, r, x...)
    bus_lookup = net._pd2pm_lookups["bus"]

    for idx in net["storage"].index:
        energy = (net["storage"].at[idx, "soc_percent"] * 1e-2 *
                  (net["storage"].at[idx, "max_e_mwh"] -
                   net["storage"].at[idx, "min_e_mwh"]))
        qs = net["storage"].at[idx, "q_mvar"].item()
        ps = net["storage"].at[idx, "p_mw"].item()
        max_p_mw = ps
        max_q_mvar, min_q_mvar = qs, -qs
        if "max_p_mw" in net["storage"]:
            max_p_mw = net["storage"].at[idx, "max_p_mw"].item()
        if "max_q_mvar" in net["storage"]:
            max_q_mvar = net["storage"].at[idx, "max_q_mvar"].item()
        if "max_q_mvar" in net["storage"]:
            min_q_mvar = net["storage"].at[idx, "min_q_mvar"].item()

        pm_idx = int(idx) + 1
        pm["storage"][str(pm_idx)] = {
            "index": pm_idx,
            "storage_bus": bus_lookup[net["storage"].at[idx, "bus"]].item(),
			"ps": ps, #* pm["baseMVA"],
            "qs": qs, #* pm["baseMVA"],            
            "energy": energy,
			"energy_rating": net["storage"].at[idx, "max_e_mwh"],
			"charge_rating": max_p_mw,
			"discharge_rating": max_p_mw,
			"charge_efficiency": 1.,
			"discharge_efficiency": 1.0,
	        "thermal_rating": net["storage"].at[idx, "max_e_mwh"], # Todo: include in DataFrame?
            "qmax": max_q_mvar,
            "qmin": min_q_mvar,
            "r": 0.0,
			"x": 0.,       
			"p_loss":-1e-8,
			"q_loss":-1e-8,			
			"status": int(net["storage"].at[idx, "in_service"]),
			"standby_loss": 0.
        }


def read_pm_storage_results(net):
    # reads the storage results from multiple time steps from the PowerModels optimization
    storage_results = dict()
    timesteps = list(net.res_ts_opt.keys())
    for idx in net.storage.index:
        # read storage results for each storage from power models to a dataframe with rows = timesteps
        res_storage = pd.DataFrame(data=None,
                                   index=timesteps,
                                   columns=["p_mw", "q_mvar", "soc_mwh", "soc_percent"],
                                   dtype=float)
        for t in timesteps:
            pm_storage = net.res_ts_opt[str(t)].res_storage
            res_storage.at[t, "p_mw"] = pm_storage["ps"]
            res_storage.at[t, "q_mvar"] = pm_storage["qs"]
            res_storage.at[t, "soc_percent"] = pm_storage["se"] * 1e2
            res_storage.at[t, "soc_mwh"] = pm_storage["se"] * \
                                           (net["storage"].at[idx, "max_e_mwh"] - net["storage"].at[idx, "min_e_mwh"])

        storage_results[idx] = res_storage

    return storage_results



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

from pandapower.create import create_bus, create_ext_grid, create_line, create_load, create_switch, create_sgen, \
    create_transformer, create_empty_network


def panda_four_load_branch():
    """
    This function creates a simple six bus system with four radial low voltage nodes connected to \
    a medium voltage slack bus. At every low voltage node the same load is connected.

    OUTPUT:
         **net** - Returns the required four load system

    EXAMPLE:
        >>> from pandapower.networks.simple_pandapower_test_networks import panda_four_load_branch
        >>> net_four_load = panda_four_load_branch()
    """
    net = create_empty_network()

    busnr1 = create_bus(net, name="bus1", vn_kv=10., geodata=(0, 0))
    busnr2 = create_bus(net, name="bus2", vn_kv=.4, geodata=(0, -1))
    busnr3 = create_bus(net, name="bus3", vn_kv=.4, geodata=(0, -2))
    busnr4 = create_bus(net, name="bus4", vn_kv=.4, geodata=(0, -3))
    busnr5 = create_bus(net, name="bus5", vn_kv=.4, geodata=(0, -4))
    busnr6 = create_bus(net, name="bus6", vn_kv=.4, geodata=(0, -5))

    create_ext_grid(net, busnr1)

    create_transformer(net, busnr1, busnr2, std_type="0.25 MVA 10/0.4 kV")

    create_line(net, busnr2, busnr3, name="line1", length_km=0.05,
                std_type="NAYY 4x120 SE")
    create_line(net, busnr3, busnr4, name="line2", length_km=0.05,
                std_type="NAYY 4x120 SE")
    create_line(net, busnr4, busnr5, name="line3", length_km=0.05,
                std_type="NAYY 4x120 SE")
    create_line(net, busnr5, busnr6, name="line4", length_km=0.05,
                std_type="NAYY 4x120 SE")

    create_load(net, busnr3, 0.030, 0.010)
    create_load(net, busnr4, 0.030, 0.010)
    create_load(net, busnr5, 0.030, 0.010)
    create_load(net, busnr6, 0.030, 0.010)
    return net


def four_loads_with_branches_out():
    """
    This function creates a simple ten bus system with four radial low voltage nodes connected to \
    a medium voltage slack bus. At every of the four radial low voltage nodes another low voltage \
    node with a load is connected via cable.

    OUTPUT:
         **net** - Returns the required four load system with branches

    EXAMPLE:
        >>> from pandapower.networks.simple_pandapower_test_networks import four_loads_with_branches_out
        >>> net_four_load_with_branches = four_loads_with_branches_out()
    """
    net = create_empty_network()

    busnr1 = create_bus(net, name="bus1ref", vn_kv=10., geodata=(0, 0))
    create_ext_grid(net, busnr1)
    busnr2 = create_bus(net, name="bus2", vn_kv=.4, geodata=(0, -1))
    create_transformer(net, busnr1, busnr2, std_type="0.25 MVA 10/0.4 kV")
    busnr3 = create_bus(net, name="bus3", vn_kv=.4, geodata=(0, -2))
    create_line(net, busnr2, busnr3, name="line1", length_km=0.05,
                std_type="NAYY 4x120 SE")
    busnr4 = create_bus(net, name="bus4", vn_kv=.4, geodata=(0, -3))
    create_line(net, busnr3, busnr4, name="line2", length_km=0.05,
                std_type="NAYY 4x120 SE")
    busnr5 = create_bus(net, name="bus5", vn_kv=.4, geodata=(0, -4))
    create_line(net, busnr4, busnr5, name="line3", length_km=0.05,
                std_type="NAYY 4x120 SE")
    busnr6 = create_bus(net, name="bus6", vn_kv=.4, geodata=(0, -5))
    create_line(net, busnr5, busnr6, name="line4", length_km=0.05,
                std_type="NAYY 4x120 SE")
    busnr7 = create_bus(net, name="bus7", vn_kv=.4, geodata=(1, -3))
    create_line(net, busnr3, busnr7, name="line5", length_km=0.05,
                std_type="NAYY 4x120 SE")
    busnr8 = create_bus(net, name="bus8", vn_kv=.4, geodata=(1, -4))
    create_line(net, busnr4, busnr8, name="line6", length_km=0.05,
                std_type="NAYY 4x120 SE")
    busnr9 = create_bus(net, name="bus9", vn_kv=.4, geodata=(1, -5))
    create_line(net, busnr5, busnr9, name="line7", length_km=0.05,
                std_type="NAYY 4x120 SE")
    busnr10 = create_bus(net, name="bus10", vn_kv=.4, geodata=(1, -6))
    create_line(net, busnr6, busnr10, name="line8", length_km=0.05,
                std_type="NAYY 4x120 SE")

    create_load(net, busnr7, p_mw=0.030, q_mvar=0.010)
    create_load(net, busnr8, p_mw=0.030, q_mvar=0.010)
    create_load(net, busnr9, p_mw=0.030, q_mvar=0.010)
    create_load(net, busnr10, p_mw=0.030, q_mvar=0.010)
    return net


def simple_four_bus_system():
    """
    This function creates a simple four bus system with two radial low voltage nodes connected to \
    a medium voltage slack bus. At both low voltage nodes the a load and a static generator is \
    connected.

    OUTPUT:
         **net** - Returns the required four bus system

    EXAMPLE:
        >>> from pandapower.networks.simple_pandapower_test_networks import simple_four_bus_system
        >>> net_simple_four_bus = simple_four_bus_system()
    """
    net = create_empty_network()
    busnr1 = create_bus(net, name="bus1ref", vn_kv=10, geodata=(0, 0))
    create_ext_grid(net, busnr1)
    busnr2 = create_bus(net, name="bus2", vn_kv=.4, geodata=(0, -1))
    create_transformer(net, busnr1, busnr2, name="transformer", std_type="0.25 MVA 10/0.4 kV")
    busnr3 = create_bus(net, name="bus3", vn_kv=.4, geodata=(0, -2))
    create_line(net, busnr2, busnr3, name="line1", length_km=0.50000, std_type="NAYY 4x50 SE")
    busnr4 = create_bus(net, name="bus4", vn_kv=.4, geodata=(0, -3))
    create_line(net, busnr3, busnr4, name="line2", length_km=0.50000, std_type="NAYY 4x50 SE")
    create_load(net, busnr3, 0.030, 0.010, name="load1")
    create_load(net, busnr4, 0.030, 0.010, name="load2")
    create_sgen(net, busnr3, p_mw=0.020, q_mvar=0.005, name="pv1", sn_mva=0.03)
    create_sgen(net, busnr4, p_mw=0.015, q_mvar=0.002, name="pv2", sn_mva=0.02)
    return net


def simple_mv_open_ring_net():
    """
    This function creates a simple medium voltage open ring network with loads at every medium \
    voltage node.
    As an example this function is used in the topology and diagnostic docu.

    OUTPUT:
         **net** - Returns the required simple medium voltage open ring network

    EXAMPLE:
         >>> from pandapower.networks.simple_pandapower_test_networks import simple_mv_open_ring_net
         >>> net_simple_open_ring = simple_mv_open_ring_net()
    """

    net = create_empty_network()

    create_bus(net, name="110 kV bar", vn_kv=110, type='b', geodata=(0, 0))
    create_bus(net, name="20 kV bar", vn_kv=20, type='b', geodata=(0, -1))
    create_bus(net, name="bus 2", vn_kv=20, type='b', geodata=(-0.5, -2))
    create_bus(net, name="bus 3", vn_kv=20, type='b', geodata=(-0.5, -3))
    create_bus(net, name="bus 4", vn_kv=20, type='b', geodata=(-0.5, -4))
    create_bus(net, name="bus 5", vn_kv=20, type='b', geodata=(0.5, -4))
    create_bus(net, name="bus 6", vn_kv=20, type='b', geodata=(0.5, -3))

    create_ext_grid(net, 0, vm_pu=1)

    create_line(net, name="line 0", from_bus=1, to_bus=2, length_km=1,
                std_type="NA2XS2Y 1x185 RM/25 12/20 kV")
    create_line(net, name="line 1", from_bus=2, to_bus=3, length_km=1,
                std_type="NA2XS2Y 1x185 RM/25 12/20 kV")
    create_line(net, name="line 2", from_bus=3, to_bus=4, length_km=1,
                std_type="NA2XS2Y 1x185 RM/25 12/20 kV")
    create_line(net, name="line 3", from_bus=4, to_bus=5, length_km=1,
                std_type="NA2XS2Y 1x185 RM/25 12/20 kV")
    create_line(net, name="line 4", from_bus=5, to_bus=6, length_km=1,
                std_type="NA2XS2Y 1x185 RM/25 12/20 kV")
    create_line(net, name="line 5", from_bus=6, to_bus=1, length_km=1,
                std_type="NA2XS2Y 1x185 RM/25 12/20 kV")

    create_transformer(net, hv_bus=0, lv_bus=1, std_type="25 MVA 110/20 kV")

    create_load(net, 2, p_mw=1, q_mvar=0.200, name="load 0")
    create_load(net, 3, p_mw=1, q_mvar=0.200, name="load 1")
    create_load(net, 4, p_mw=1, q_mvar=0.200, name="load 2")
    create_load(net, 5, p_mw=1, q_mvar=0.200, name="load 3")
    create_load(net, 6, p_mw=1, q_mvar=0.200, name="load 4")

    create_switch(net, bus=1, element=0, et='l')
    create_switch(net, bus=2, element=0, et='l')
    create_switch(net, bus=2, element=1, et='l')
    create_switch(net, bus=3, element=1, et='l')
    create_switch(net, bus=3, element=2, et='l')
    create_switch(net, bus=4, element=2, et='l')
    create_switch(net, bus=4, element=3, et='l', closed=0)
    create_switch(net, bus=5, element=3, et='l')
    create_switch(net, bus=5, element=4, et='l')
    create_switch(net, bus=6, element=4, et='l')
    create_switch(net, bus=6, element=5, et='l')
    create_switch(net, bus=1, element=5, et='l')
    return net



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

from numpy import nan

from pandapower.create import create_empty_network, create_bus, create_line, create_load, \
    create_transformer_from_parameters, create_shunt, create_gen, create_ext_grid, create_buses, create_switch, \
    create_load_from_cosphi, create_sgen, create_storage
from pandapower.std_types import create_std_type

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging
logger = logging.getLogger(__name__)


def create_cigre_network_hv(length_km_6a_6b=0.1):
    """
    Create the CIGRE HV Grid from final Report of Task Force C6.04.02:
    "Benchmark Systems for Network Integration of Renewable and Distributed Energy Resources”, 2014.

    OPTIONAL:
        **length_km_6a_6b** (float, 0.1) - Length of the line segment 9 between buses 6a and 6b
            which is intended to be optional with user-definable geometrical configuration, length
            and installation type.

    OUTPUT:
        **net** - The pandapower format network.
    """
    net_cigre_hv = create_empty_network()

    # Linedata
    # Line220kV
    line_data = {'c_nf_per_km': 9.08, 'r_ohm_per_km': 0.0653,
                 'x_ohm_per_km': 0.398, 'max_i_ka': 1.14,
                 'type': 'ol'}

    create_std_type(net_cigre_hv, line_data, 'Line220kV', element='line')

    # Line380kV
    line_data = {'c_nf_per_km': 11.5, 'r_ohm_per_km': 0.0328,
                 'x_ohm_per_km': 0.312, 'max_i_ka': 1.32,
                 'type': 'ol'}

    create_std_type(net_cigre_hv, line_data, 'Line380kV', element='line')

    # Busses
    bus1 = create_bus(net_cigre_hv, name='Bus 1', vn_kv=220, type='b', zone='CIGRE_HV')
    bus2 = create_bus(net_cigre_hv, name='Bus 2', vn_kv=220, type='b', zone='CIGRE_HV')
    bus3 = create_bus(net_cigre_hv, name='Bus 3', vn_kv=220, type='b', zone='CIGRE_HV')
    bus4 = create_bus(net_cigre_hv, name='Bus 4', vn_kv=220, type='b', zone='CIGRE_HV')
    bus5 = create_bus(net_cigre_hv, name='Bus 5', vn_kv=220, type='b', zone='CIGRE_HV')
    bus6a = create_bus(net_cigre_hv, name='Bus 6a', vn_kv=220, type='b', zone='CIGRE_HV')
    bus6b = create_bus(net_cigre_hv, name='Bus 6b', vn_kv=220, type='b', zone='CIGRE_HV')
    bus7 = create_bus(net_cigre_hv, name='Bus 7', vn_kv=380, type='b', zone='CIGRE_HV')
    bus8 = create_bus(net_cigre_hv, name='Bus 8', vn_kv=380, type='b', zone='CIGRE_HV')
    bus9 = create_bus(net_cigre_hv, name='Bus 9', vn_kv=22, type='b', zone='CIGRE_HV')
    bus10 = create_bus(net_cigre_hv, name='Bus 10', vn_kv=22, type='b', zone='CIGRE_HV')
    bus11 = create_bus(net_cigre_hv, name='Bus 11', vn_kv=22, type='b', zone='CIGRE_HV')
    bus12 = create_bus(net_cigre_hv, name='Bus 12', vn_kv=22, type='b', zone='CIGRE_HV')

    # Lines
    create_line(net_cigre_hv, bus1, bus2, length_km=100, std_type='Line220kV', name='Line 1-2')
    create_line(net_cigre_hv, bus1, bus6a, length_km=300, std_type='Line220kV', name='Line 1-6a')
    create_line(net_cigre_hv, bus2, bus5, length_km=300, std_type='Line220kV', name='Line 2-5')
    create_line(net_cigre_hv, bus3, bus4, length_km=100, std_type='Line220kV', name='Line 3-4')
    create_line(net_cigre_hv, bus3, bus4, length_km=100, std_type='Line220kV', name='Line 3-4_2')
    create_line(net_cigre_hv, bus4, bus5, length_km=300, std_type='Line220kV', name='Line 4-5')
    create_line(net_cigre_hv, bus4, bus6a, length_km=300, std_type='Line220kV', name='Line 4-6a')
    create_line(net_cigre_hv, bus7, bus8, length_km=600, std_type='Line380kV', name='Line 7-8')
    create_line(net_cigre_hv, bus6a, bus6b, length_km=length_km_6a_6b, std_type='Line220kV', name='Line 6a-6b')

    # Trafos
    create_transformer_from_parameters(net_cigre_hv, bus7, bus1, sn_mva=1000,
                                       vn_hv_kv=380, vn_lv_kv=220, vkr_percent=0.0,
                                       vk_percent=13.0, pfe_kw=0, i0_percent=0,
                                       shift_degree=0.0, name='Trafo 1-7')
    create_transformer_from_parameters(net_cigre_hv, bus8, bus3, sn_mva=1000,
                                       vn_hv_kv=380, vn_lv_kv=220, vkr_percent=0.0,
                                       vk_percent=13.0, pfe_kw=0, i0_percent=0,
                                       shift_degree=0.0, name='Trafo 3-8')

    create_transformer_from_parameters(net_cigre_hv, bus1, bus9, sn_mva=1000,
                                       vn_hv_kv=220, vn_lv_kv=22, vkr_percent=0.0,
                                       vk_percent=13.0, pfe_kw=0, i0_percent=0,
                                       shift_degree=330.0, name='Trafo 9-1')
    create_transformer_from_parameters(net_cigre_hv, bus2, bus10, sn_mva=1000,
                                       vn_hv_kv=220, vn_lv_kv=22, vkr_percent=0.0,
                                       vk_percent=13.0, pfe_kw=0, i0_percent=0,
                                       shift_degree=330.0, name='Trafo 10-2')
    create_transformer_from_parameters(net_cigre_hv, bus3, bus11, sn_mva=1000,
                                       vn_hv_kv=220, vn_lv_kv=22, vkr_percent=0.0,
                                       vk_percent=13.0, pfe_kw=0, i0_percent=0,
                                       shift_degree=330.0, name='Trafo 11-3')
    create_transformer_from_parameters(net_cigre_hv, bus6b, bus12, sn_mva=500,
                                       vn_hv_kv=220, vn_lv_kv=22, vkr_percent=0.0,
                                       vk_percent=13.0, pfe_kw=0, i0_percent=0,
                                       shift_degree=330.0, name='Trafo 12-6b')

    # Loads
    create_load(net_cigre_hv, bus2, p_mw=285, q_mvar=200, name='Load 2')
    create_load(net_cigre_hv, bus3, p_mw=325, q_mvar=244, name='Load 3')
    create_load(net_cigre_hv, bus4, p_mw=326, q_mvar=244, name='Load 4')
    create_load(net_cigre_hv, bus5, p_mw=103, q_mvar=62, name='Load 5')
    create_load(net_cigre_hv, bus6a, p_mw=435, q_mvar=296, name='Load 6a')

    # External grid
    create_ext_grid(net_cigre_hv, bus9, vm_pu=1.03, va_degree=0, name='Generator 9')

    # Generators
    create_gen(net_cigre_hv, bus10, vm_pu=1.03, p_mw=500, name='Generator 10')
    create_gen(net_cigre_hv, bus11, vm_pu=1.03, p_mw=200, name='Generator 11')
    create_gen(net_cigre_hv, bus12, vm_pu=1.03, p_mw=300, name='Generator 12')

    # Shunts
    create_shunt(net_cigre_hv, bus4, p_mw=0.0, q_mvar=-160, name='Shunt 4')
    create_shunt(net_cigre_hv, bus5, p_mw=0.0, q_mvar=-80, name='Shunt 5')
    create_shunt(net_cigre_hv, bus6a, p_mw=0.0, q_mvar=-180, name='Shunt 6a')

    # Bus geo data
    bus_geodata = list(map(lambda xy: f'{{"type":"Point", "coordinates":[{xy[0]}, {xy[1]}]}}', zip(
        [4, 8, 20, 16, 12, 8, 12, 4, 20, 0, 8, 24, 16],
        [8.0, 8.0, 8.0, 8.0, 8.0, 6.0, 4.5, 1.0, 1.0, 8.0, 12.0, 8.0, 4.5])))
    # Match bus.index
    net_cigre_hv.bus["geo"] = bus_geodata

    return net_cigre_hv


def create_cigre_network_mv(with_der=False):
    """
    Create the CIGRE MV Grid from final Report of Task Force C6.04.02:
    "Benchmark Systems for Network Integration of Renewable and Distributed Energy Resources”, 2014.

    OPTIONAL:
        **with_der** (boolean or str, False) - Range of DER consideration, which should be in
            (False, "pv_wind", "all"). The DER types, dimensions and locations are taken from CIGRE
            CaseStudy: "DER in Medium Voltage Systems"

    OUTPUT:
        **net** - The pandapower format network.
    """
    if with_der is True:
        raise ValueError("'with_der=True' is deprecated. Please use 'with_der=pv_wind'")
    if with_der not in [False, "pv_wind", "all"]:
        raise ValueError("'with_der' is unknown. It should be in [False, 'pv_wind', 'all'].")

    net_cigre_mv = create_empty_network()

    # Linedata
    line_data = {'c_nf_per_km': 151.1749, 'r_ohm_per_km': 0.501,
                 'x_ohm_per_km': 0.716, 'max_i_ka': 0.145,
                 'type': 'cs'}
    create_std_type(net_cigre_mv, line_data, name='CABLE_CIGRE_MV', element='line')

    line_data = {'c_nf_per_km': 10.09679, 'r_ohm_per_km': 0.510,
                 'x_ohm_per_km': 0.366, 'max_i_ka': 0.195,
                 'type': 'ol'}
    create_std_type(net_cigre_mv, line_data, name='OHL_CIGRE_MV', element='line')

    # Busses
    bus_geodata = list(zip([7., 4., 4., 4., 2.5, 1., 1., 8., 8., 6., 4., 4., 10., 10., 10.],
                           [16., 15., 13., 11., 9., 7., 3., 3., 5., 5., 5., 7., 15., 11., 5.]))
    bus0 = create_bus(net_cigre_mv, name='Bus 0', vn_kv=110, type='b', zone='CIGRE_MV', geodata=bus_geodata[0])
    buses = create_buses(net_cigre_mv, 14, name=['Bus %i' % i for i in range(1, 15)], vn_kv=20,
                         type='b', zone='CIGRE_MV', geodata=bus_geodata[1:])

    # Lines
    create_line(net_cigre_mv, buses[0], buses[1], length_km=2.82, std_type='CABLE_CIGRE_MV', name='Line 1-2')
    create_line(net_cigre_mv, buses[1], buses[2], length_km=4.42, std_type='CABLE_CIGRE_MV', name='Line 2-3')
    create_line(net_cigre_mv, buses[2], buses[3], length_km=0.61, std_type='CABLE_CIGRE_MV', name='Line 3-4')
    create_line(net_cigre_mv, buses[3], buses[4], length_km=0.56, std_type='CABLE_CIGRE_MV', name='Line 4-5')
    create_line(net_cigre_mv, buses[4], buses[5], length_km=1.54, std_type='CABLE_CIGRE_MV', name='Line 5-6')
    create_line(net_cigre_mv, buses[6], buses[7], length_km=1.67, std_type='CABLE_CIGRE_MV', name='Line 7-8')
    create_line(net_cigre_mv, buses[7], buses[8], length_km=0.32, std_type='CABLE_CIGRE_MV', name='Line 8-9')
    create_line(net_cigre_mv, buses[8], buses[9], length_km=0.77, std_type='CABLE_CIGRE_MV', name='Line 9-10')
    create_line(net_cigre_mv, buses[9], buses[10], length_km=0.33, std_type='CABLE_CIGRE_MV', name='Line 10-11')
    create_line(net_cigre_mv, buses[2], buses[7], length_km=1.3, std_type='CABLE_CIGRE_MV', name='Line 3-8')
    create_line(net_cigre_mv, buses[11], buses[12], length_km=4.89, std_type='OHL_CIGRE_MV', name='Line 12-13')
    create_line(net_cigre_mv, buses[12], buses[13], length_km=2.99, std_type='OHL_CIGRE_MV', name='Line 13-14')

    line6_7 = create_line(net_cigre_mv, buses[5], buses[6], length_km=0.24, std_type='CABLE_CIGRE_MV', name='Line 6-7')
    line4_11 = create_line(net_cigre_mv, buses[10], buses[3], length_km=0.49, std_type='CABLE_CIGRE_MV',
                           name='Line 11-4')
    line8_14 = create_line(net_cigre_mv, buses[13], buses[7], length_km=2., std_type='OHL_CIGRE_MV', name='Line 14-8')

    # Ext-Grid
    create_ext_grid(
        net_cigre_mv, bus0, vm_pu=1.03, va_degree=0., s_sc_max_mva=5000, s_sc_min_mva=5000, rx_max=0.1, rx_min=0.1
    )

    # Trafos
    trafo0 = create_transformer_from_parameters(net_cigre_mv, bus0, buses[0], sn_mva=25,
                                                vn_hv_kv=110, vn_lv_kv=20, vkr_percent=0.16,
                                                vk_percent=12.00107, pfe_kw=0, i0_percent=0,
                                                shift_degree=30.0, name='Trafo 0-1')
    trafo1 = create_transformer_from_parameters(net_cigre_mv, bus0, buses[11], sn_mva=25,
                                                vn_hv_kv=110, vn_lv_kv=20, vkr_percent=0.16,
                                                vk_percent=12.00107, pfe_kw=0, i0_percent=0,
                                                shift_degree=30.0, name='Trafo 0-12')

    # Switches
    # S2
    create_switch(net_cigre_mv, buses[5], line6_7, et='l', closed=True, type='LBS')
    create_switch(net_cigre_mv, buses[6], line6_7, et='l', closed=False, type='LBS', name='S2')
    # S3
    create_switch(net_cigre_mv, buses[3], line4_11, et='l', closed=False, type='LBS', name='S3')
    create_switch(net_cigre_mv, buses[10], line4_11, et='l', closed=True, type='LBS')
    # S1
    create_switch(net_cigre_mv, buses[7], line8_14, et='l', closed=False, type='LBS', name='S1')
    create_switch(net_cigre_mv, buses[13], line8_14, et='l', closed=True, type='LBS')
    # trafos
    create_switch(net_cigre_mv, bus0, trafo0, et='t', closed=True, type='CB')
    create_switch(net_cigre_mv, bus0, trafo1, et='t', closed=True, type='CB')

    # Loads
    # Residential
    create_load_from_cosphi(net_cigre_mv, buses[0], 15.3, 0.98, "underexcited", name='Load R1')
    create_load_from_cosphi(net_cigre_mv, buses[2], 0.285, 0.97, "underexcited", name='Load R3')
    create_load_from_cosphi(net_cigre_mv, buses[3], 0.445, 0.97, "underexcited", name='Load R4')
    create_load_from_cosphi(net_cigre_mv, buses[4], 0.750, 0.97, "underexcited", name='Load R5')
    create_load_from_cosphi(net_cigre_mv, buses[5], 0.565, 0.97, "underexcited", name='Load R6')
    create_load_from_cosphi(net_cigre_mv, buses[7], 0.605, 0.97, "underexcited", name='Load R8')
    create_load_from_cosphi(net_cigre_mv, buses[9], 0.490, 0.97, "underexcited", name='Load R10')
    create_load_from_cosphi(net_cigre_mv, buses[10], 0.340, 0.97, "underexcited", name='Load R11')
    create_load_from_cosphi(net_cigre_mv, buses[11], 15.3, 0.98, "underexcited", name='Load R12')
    create_load_from_cosphi(net_cigre_mv, buses[13], 0.215, 0.97, "underexcited", name='Load R14')

    # Commercial / Industrial
    create_load_from_cosphi(net_cigre_mv, buses[0], 5.1, 0.95, "underexcited", name='Load CI1')
    create_load_from_cosphi(net_cigre_mv, buses[2], 0.265, 0.85, "underexcited", name='Load CI3')
    create_load_from_cosphi(net_cigre_mv, buses[6], 0.090, 0.85, "underexcited", name='Load CI7')
    create_load_from_cosphi(net_cigre_mv, buses[8], 0.675, 0.85, "underexcited", name='Load CI9')
    create_load_from_cosphi(net_cigre_mv, buses[9], 0.080, 0.85, "underexcited", name='Load CI10')
    create_load_from_cosphi(net_cigre_mv, buses[11], 5.28, 0.95, "underexcited", name='Load CI12')
    create_load_from_cosphi(net_cigre_mv, buses[12], 0.04, 0.85, "underexcited", name='Load CI13')
    create_load_from_cosphi(net_cigre_mv, buses[13], 0.390, 0.85, "underexcited", name='Load CI14')

    # Optional distributed energy recources
    if with_der in ["pv_wind", "all"]:
        create_sgen(net_cigre_mv, buses[2], 0.02, q_mvar=0, sn_mva=0.02, name='PV 3', type='PV')
        create_sgen(net_cigre_mv, buses[3], 0.020, q_mvar=0, sn_mva=0.02, name='PV 4', type='PV')
        create_sgen(net_cigre_mv, buses[4], 0.030, q_mvar=0, sn_mva=0.03, name='PV 5', type='PV')
        create_sgen(net_cigre_mv, buses[5], 0.030, q_mvar=0, sn_mva=0.03, name='PV 6', type='PV')
        create_sgen(net_cigre_mv, buses[7], 0.030, q_mvar=0, sn_mva=0.03, name='PV 8', type='PV')
        create_sgen(net_cigre_mv, buses[8], 0.030, q_mvar=0, sn_mva=0.03, name='PV 9', type='PV')
        create_sgen(net_cigre_mv, buses[9], 0.040, q_mvar=0, sn_mva=0.04, name='PV 10', type='PV')
        create_sgen(net_cigre_mv, buses[10], 0.010, q_mvar=0, sn_mva=0.01, name='PV 11', type='PV')
        create_sgen(net_cigre_mv, buses[6], 1.5, q_mvar=0, sn_mva=1.5, name='WKA 7',
                    type='WP')
        if with_der == "all":
            create_storage(net_cigre_mv, bus=buses[4], p_mw=0.6, max_e_mwh=nan, sn_mva=0.2,
                           name='Battery 1', type='Battery', max_p_mw=0.6, min_p_mw=-0.6)
            create_sgen(net_cigre_mv, bus=buses[4], p_mw=0.033, sn_mva=0.033,
                        name='Residential fuel cell 1', type='Residential fuel cell')
            create_sgen(net_cigre_mv, bus=buses[8], p_mw=0.310, sn_mva=0.31, name='CHP diesel 1',
                        type='CHP diesel')
            create_sgen(net_cigre_mv, bus=buses[8], p_mw=0.212, sn_mva=0.212, name='Fuel cell 1',
                        type='Fuel cell')
            create_storage(net_cigre_mv, bus=buses[9], p_mw=0.200, max_e_mwh=nan, sn_mva=0.2,
                           name='Battery 2', type='Battery', max_p_mw=0.2, min_p_mw=-0.2)
            create_sgen(net_cigre_mv, bus=buses[9], p_mw=0.014, sn_mva=.014,
                        name='Residential fuel cell 2', type='Residential fuel cell')

    return net_cigre_mv


def create_cigre_network_lv():
    """
    Create the CIGRE LV Grid from final Report of Task Force C6.04.02:
    "Benchmark Systems for Network Integration of Renewable and Distributed Energy Resources”, 2014.

    OUTPUT:
        **net** - The pandapower format network.
    """
    net_cigre_lv = create_empty_network()

    # Linedata
    # UG1
    line_data = {'c_nf_per_km': 0.0, 'r_ohm_per_km': 0.162,
                 'x_ohm_per_km': 0.0832, 'max_i_ka': 1.0,
                 'type': 'cs'}
    create_std_type(net_cigre_lv, line_data, name='UG1', element='line')

    # UG2
    line_data = {'c_nf_per_km': 0.0, 'r_ohm_per_km': 0.2647,
                 'x_ohm_per_km': 0.0823, 'max_i_ka': 1.0,
                 'type': 'cs'}
    create_std_type(net_cigre_lv, line_data, name='UG2', element='line')

    # UG3
    line_data = {'c_nf_per_km': 0.0, 'r_ohm_per_km': 0.822,
                 'x_ohm_per_km': 0.0847, 'max_i_ka': 1.0,
                 'type': 'cs'}
    create_std_type(net_cigre_lv, line_data, name='UG3', element='line')

    # OH1
    line_data = {'c_nf_per_km': 0.0, 'r_ohm_per_km': 0.4917,
                 'x_ohm_per_km': 0.2847, 'max_i_ka': 1.0,
                 'type': 'ol'}
    create_std_type(net_cigre_lv, line_data, name='OH1', element='line')

    # OH2
    line_data = {'c_nf_per_km': 0.0, 'r_ohm_per_km': 1.3207,
                 'x_ohm_per_km': 0.321, 'max_i_ka': 1.0,
                 'type': 'ol'}
    create_std_type(net_cigre_lv, line_data, name='OH2', element='line')

    # OH3
    line_data = {'c_nf_per_km': 0.0, 'r_ohm_per_km': 2.0167,
                 'x_ohm_per_km': 0.3343, 'max_i_ka': 1.0,
                 'type': 'ol'}
    create_std_type(net_cigre_lv, line_data, name='OH3', element='line')

    # Busses
    bus0 = create_bus(net_cigre_lv, name='Bus 0', vn_kv=20.0, type='b', zone='CIGRE_LV')
    busR0 = create_bus(net_cigre_lv, name='Bus R0', vn_kv=20.0, type='b', zone='CIGRE_LV')
    busR1 = create_bus(net_cigre_lv, name='Bus R1', vn_kv=0.4, type='b', zone='CIGRE_LV')
    busR2 = create_bus(net_cigre_lv, name='Bus R2', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busR3 = create_bus(net_cigre_lv, name='Bus R3', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busR4 = create_bus(net_cigre_lv, name='Bus R4', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busR5 = create_bus(net_cigre_lv, name='Bus R5', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busR6 = create_bus(net_cigre_lv, name='Bus R6', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busR7 = create_bus(net_cigre_lv, name='Bus R7', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busR8 = create_bus(net_cigre_lv, name='Bus R8', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busR9 = create_bus(net_cigre_lv, name='Bus R9', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busR10 = create_bus(net_cigre_lv, name='Bus R10', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busR11 = create_bus(net_cigre_lv, name='Bus R11', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busR12 = create_bus(net_cigre_lv, name='Bus R12', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busR13 = create_bus(net_cigre_lv, name='Bus R13', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busR14 = create_bus(net_cigre_lv, name='Bus R14', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busR15 = create_bus(net_cigre_lv, name='Bus R15', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busR16 = create_bus(net_cigre_lv, name='Bus R16', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busR17 = create_bus(net_cigre_lv, name='Bus R17', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busR18 = create_bus(net_cigre_lv, name='Bus R18', vn_kv=0.4, type='m', zone='CIGRE_LV')

    busI0 = create_bus(net_cigre_lv, name='Bus I0', vn_kv=20.0, type='b', zone='CIGRE_LV')
    busI1 = create_bus(net_cigre_lv, name='Bus I1', vn_kv=0.4, type='b', zone='CIGRE_LV')
    busI2 = create_bus(net_cigre_lv, name='Bus I2', vn_kv=0.4, type='m', zone='CIGRE_LV')

    busC0 = create_bus(net_cigre_lv, name='Bus C0', vn_kv=20.0, type='b', zone='CIGRE_LV')
    busC1 = create_bus(net_cigre_lv, name='Bus C1', vn_kv=0.4, type='b', zone='CIGRE_LV')
    busC2 = create_bus(net_cigre_lv, name='Bus C2', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busC3 = create_bus(net_cigre_lv, name='Bus C3', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busC4 = create_bus(net_cigre_lv, name='Bus C4', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busC5 = create_bus(net_cigre_lv, name='Bus C5', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busC6 = create_bus(net_cigre_lv, name='Bus C6', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busC7 = create_bus(net_cigre_lv, name='Bus C7', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busC8 = create_bus(net_cigre_lv, name='Bus C8', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busC9 = create_bus(net_cigre_lv, name='Bus C9', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busC10 = create_bus(net_cigre_lv, name='Bus C10', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busC11 = create_bus(net_cigre_lv, name='Bus C11', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busC12 = create_bus(net_cigre_lv, name='Bus C12', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busC13 = create_bus(net_cigre_lv, name='Bus C13', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busC14 = create_bus(net_cigre_lv, name='Bus C14', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busC15 = create_bus(net_cigre_lv, name='Bus C15', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busC16 = create_bus(net_cigre_lv, name='Bus C16', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busC17 = create_bus(net_cigre_lv, name='Bus C17', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busC18 = create_bus(net_cigre_lv, name='Bus C18', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busC19 = create_bus(net_cigre_lv, name='Bus C19', vn_kv=0.4, type='m', zone='CIGRE_LV')
    busC20 = create_bus(net_cigre_lv, name='Bus C20', vn_kv=0.4, type='m', zone='CIGRE_LV')

    # Lines
    create_line(net_cigre_lv, busR1, busR2, length_km=0.035, std_type='UG1', name='Line R1-R2')
    create_line(net_cigre_lv, busR2, busR3, length_km=0.035, std_type='UG1', name='Line R2-R3')
    create_line(net_cigre_lv, busR3, busR4, length_km=0.035, std_type='UG1', name='Line R3-R4')
    create_line(net_cigre_lv, busR4, busR5, length_km=0.035, std_type='UG1', name='Line R4-R5')
    create_line(net_cigre_lv, busR5, busR6, length_km=0.035, std_type='UG1', name='Line R5-R6')
    create_line(net_cigre_lv, busR6, busR7, length_km=0.035, std_type='UG1', name='Line R6-R7')
    create_line(net_cigre_lv, busR7, busR8, length_km=0.035, std_type='UG1', name='Line R7-R8')
    create_line(net_cigre_lv, busR8, busR9, length_km=0.035, std_type='UG1', name='Line R8-R9')
    create_line(net_cigre_lv, busR9, busR10, length_km=0.035, std_type='UG1', name='Line R9-R10')
    create_line(net_cigre_lv, busR3, busR11, length_km=0.030, std_type='UG3', name='Line R3-R11')
    create_line(net_cigre_lv, busR4, busR12, length_km=0.035, std_type='UG3', name='Line R4-R12')
    create_line(net_cigre_lv, busR12, busR13, length_km=0.035, std_type='UG3', name='Line R12-R13')
    create_line(net_cigre_lv, busR13, busR14, length_km=0.035, std_type='UG3', name='Line R13-R14')
    create_line(net_cigre_lv, busR14, busR15, length_km=0.030, std_type='UG3', name='Line R14-R15')
    create_line(net_cigre_lv, busR6, busR16, length_km=0.030, std_type='UG3', name='Line R6-R16')
    create_line(net_cigre_lv, busR9, busR17, length_km=0.030, std_type='UG3', name='Line R9-R17')
    create_line(net_cigre_lv, busR10, busR18, length_km=0.030, std_type='UG3', name='Line R10-R18')

    create_line(net_cigre_lv, busI1, busI2, length_km=0.2, std_type='UG2', name='Line I1-I2')

    create_line(net_cigre_lv, busC1, busC2, length_km=0.030, std_type='OH1', name='Line C1-C2')
    create_line(net_cigre_lv, busC2, busC3, length_km=0.030, std_type='OH1', name='Line C2-C3')
    create_line(net_cigre_lv, busC3, busC4, length_km=0.030, std_type='OH1', name='Line C3-C4')
    create_line(net_cigre_lv, busC4, busC5, length_km=0.030, std_type='OH1', name='Line C4-C5')
    create_line(net_cigre_lv, busC5, busC6, length_km=0.030, std_type='OH1', name='Line C5-C6')
    create_line(net_cigre_lv, busC6, busC7, length_km=0.030, std_type='OH1', name='Line C6-C7')
    create_line(net_cigre_lv, busC7, busC8, length_km=0.030, std_type='OH1', name='Line C7-C8')
    create_line(net_cigre_lv, busC8, busC9, length_km=0.030, std_type='OH1', name='Line C8-C9')
    create_line(net_cigre_lv, busC3, busC10, length_km=0.030, std_type='OH2', name='Line C3-C10')
    create_line(net_cigre_lv, busC10, busC11, length_km=0.030, std_type='OH2', name='Line C10-C11')
    create_line(net_cigre_lv, busC11, busC12, length_km=0.030, std_type='OH3', name='Line C11-C12')
    create_line(net_cigre_lv, busC11, busC13, length_km=0.030, std_type='OH3', name='Line C11-C13')
    create_line(net_cigre_lv, busC10, busC14, length_km=0.030, std_type='OH3', name='Line C10-C14')
    create_line(net_cigre_lv, busC5, busC15, length_km=0.030, std_type='OH2', name='Line C5-C15')
    create_line(net_cigre_lv, busC15, busC16, length_km=0.030, std_type='OH2', name='Line C15-C16')
    create_line(net_cigre_lv, busC15, busC17, length_km=0.030, std_type='OH3', name='Line C15-C17')
    create_line(net_cigre_lv, busC16, busC18, length_km=0.030, std_type='OH3', name='Line C16-C18')
    create_line(net_cigre_lv, busC8, busC19, length_km=0.030, std_type='OH3', name='Line C8-C19')
    create_line(net_cigre_lv, busC9, busC20, length_km=0.030, std_type='OH3', name='Line C9-C20')

    # Trafos
    create_transformer_from_parameters(net_cigre_lv, busR0, busR1, sn_mva=0.5, vn_hv_kv=20.0,
                                       vn_lv_kv=0.4, vkr_percent=1.0, vk_percent=4.123106,
                                       pfe_kw=0.0, i0_percent=0.0, shift_degree=30.0,
                                       tap_pos=0.0, name='Trafo R0-R1')

    create_transformer_from_parameters(net_cigre_lv, busI0, busI1, sn_mva=0.15, vn_hv_kv=20.0,
                                       vn_lv_kv=0.4, vkr_percent=1.003125, vk_percent=4.126896,
                                       pfe_kw=0.0, i0_percent=0.0, shift_degree=30.0,
                                       tap_pos=0.0, name='Trafo I0-I1')

    create_transformer_from_parameters(net_cigre_lv, busC0, busC1, sn_mva=0.3, vn_hv_kv=20.0,
                                       vn_lv_kv=0.4, vkr_percent=0.993750, vk_percent=4.115529,
                                       pfe_kw=0.0, i0_percent=0.0, shift_degree=30.0,
                                       tap_pos=0.0, name='Trafo C0-C1')

    # External grid
    create_ext_grid(net_cigre_lv, bus0, vm_pu=1.0, va_degree=0.0, s_sc_max_mva=100.0,
                    s_sc_min_mva=100.0, rx_max=1.0, rx_min=1.0)

    # Loads
    create_load(net_cigre_lv, busR1, p_mw=0.19, q_mvar=0.062449980, name='Load R1')
    create_load(net_cigre_lv, busR11, p_mw=0.01425, q_mvar=0.004683748, name='Load R11')
    create_load(net_cigre_lv, busR15, p_mw=0.0494, q_mvar=0.016236995, name='Load R15')
    create_load(net_cigre_lv, busR16, p_mw=0.05225, q_mvar=0.017173744, name='Load R16')
    create_load(net_cigre_lv, busR17, p_mw=0.03325, q_mvar=0.010928746, name='Load R17')
    create_load(net_cigre_lv, busR18, p_mw=0.04465, q_mvar=0.014675745, name='Load R18')
    create_load(net_cigre_lv, busI2, p_mw=0.0850, q_mvar=0.052678269, name='Load I2')
    create_load(net_cigre_lv, busC1, p_mw=0.1080, q_mvar=0.052306787, name='Load C1')
    create_load(net_cigre_lv, busC12, p_mw=0.018, q_mvar=0.008717798, name='Load C12')
    create_load(net_cigre_lv, busC13, p_mw=0.018, q_mvar=0.008717798, name='Load C13')
    create_load(net_cigre_lv, busC14, p_mw=0.0225, q_mvar=0.010897247, name='Load C14')
    create_load(net_cigre_lv, busC17, p_mw=0.0225, q_mvar=0.010897247, name='Load C17')
    create_load(net_cigre_lv, busC18, p_mw=0.0072, q_mvar=0.003487119, name='Load C18')
    create_load(net_cigre_lv, busC19, p_mw=0.0144, q_mvar=0.006974238, name='Load C19')
    create_load(net_cigre_lv, busC20, p_mw=0.0072, q_mvar=0.003487119, name='Load C20')

    # Switches
    create_switch(net_cigre_lv, bus0, busR0, et='b', closed=True, type='CB', name='S1')
    create_switch(net_cigre_lv, bus0, busI0, et='b', closed=True, type='CB', name='S2')
    create_switch(net_cigre_lv, bus0, busC0, et='b', closed=True, type='CB', name='S3')

    # Bus geo data
    bus_geodata = list(map(lambda xy: f'{{"type":"Point", "coordinates":[{xy[0]}, {xy[1]}]}}', zip(
        [0.2, 0.2, -1.4583333333, -1.4583333333, -1.4583333333, -1.9583333333, -2.7083333333, -2.7083333333,
         -3.2083333333, -3.2083333333, -3.2083333333, -3.7083333333, -0.9583333333, -1.2083333333, -1.2083333333,
         -1.2083333333, -1.2083333333, -2.2083333333, -2.7083333333, -3.7083333333, 0.2, 0.2, 0.2, 0.2, 1.9166666667,
         1.9166666667, 1.9166666667, 0.5416666667, 0.5416666667, -0.2083333333, -0.2083333333, -0.2083333333,
         -0.7083333333, 3.2916666667, 2.7916666667, 2.2916666667, 3.2916666667, 3.7916666667, 1.2916666667,
         0.7916666667, 1.7916666667, 0.7916666667, 0.2916666667, -0.7083333333],
        [1.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 11.0, 5.0, 6.0, 7.0, 8.0, 9.0, 8.0, 11.0,
         12.0, 1.0, 2.0, 3.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 5.0, 6.0, 7.0, 7.0, 6.0,
         7.0, 8.0, 8.0, 9.0, 10.0, 11.0])))
    # Match bus.index
    net_cigre_lv.bus["geo"] = bus_geodata
    return net_cigre_lv


if __name__ == "__main__":
    net = create_cigre_network_lv()
    net = create_cigre_network_mv()
    net = create_cigre_network_hv()



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import tempfile
from collections.abc import Iterable

import tqdm

from pandapower.auxiliary import ControllerNotConverged
from pandapower.control import prepare_run_ctrl, run_control
from pandapower.control.util.diagnostic import control_diagnostic
from pandapower.run import runpp
from pandapower.timeseries.output_writer import OutputWriter

try:
    import pandaplan.core.pplog as pplog
except ImportError:
    import logging as pplog

logger = pplog.getLogger(__name__)
logger.setLevel(level=pplog.WARNING)


def init_default_outputwriter(net, time_steps, **kwargs):
    """
    Initializes the output writer. If output_writer is None, default output_writer is created

    INPUT:
        **net** - The pandapower format network

        **time_steps** (list) - time steps to be calculated

    """
    output_writer = kwargs.get("output_writer", None)
    if output_writer is not None:
        # write the output_writer to net
        logger.warning("deprecated: output_writer should not be given to run_timeseries(). "
                       "This overwrites the stored one in net.output_writer.")
        net.output_writer.iat[0, 0] = output_writer
    if "output_writer" not in net or net.output_writer.iat[0, 0] is None:
        # create a default output writer for this net
        ow = OutputWriter(net, time_steps, output_path=tempfile.gettempdir())
        logger.info("No output writer specified. Using default:")
        logger.info(ow)


def init_output_writer(net, time_steps):
    # init output writer before time series calculation
    output_writer = net.output_writer.iat[0, 0]
    output_writer.time_steps = time_steps
    output_writer.init_all(net)


#
# def print_progress_bar(iteration, total, prefix='', suffix='', decimals=1, length=100, fill='█'):
#     """
#     Call in a loop to create terminal progress bar.
#     the code is mentioned in : https://stackoverflow.com/questions/3173320/text-progress-bar-in-the-console
#     """
#     percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
#     filled_length = int(length * iteration // total)
#     bar = fill * filled_length + '-' * (length - filled_length)
#     # logger.info('\r%s |%s| %s%% %s' % (prefix, bar, percent, suffix))
#     print('\r%s |%s| %s%% %s' % (prefix, bar, percent, suffix), end="")
#     # Print New Line on Complete
#     if iteration == total:
#         print("\n")


def controller_not_converged(time_step, ts_variables):
    logger.error('ControllerNotConverged at time step %s' % time_step)
    if not ts_variables["continue_on_divergence"]:
        raise ControllerNotConverged


def pf_not_converged(time_step, ts_variables):
    logger.error('CalculationNotConverged at time step %s' % time_step)
    if not ts_variables["continue_on_divergence"]:
        raise ts_variables['errors'][0]


def control_time_step(controller_order, time_step):
    for levelorder in controller_order:
        for ctrl, net in levelorder:
            ctrl.time_step(net, time_step)


def finalize_step(controller_order, time_step):
    for levelorder in controller_order:
        for ctrl, net in levelorder:
            ctrl.finalize_step(net, time_step)


def output_writer_routine(net, time_step, pf_converged, ctrl_converged, recycle_options):
    output_writer = net["output_writer"].iat[0, 0]
    # update time step for output writer
    output_writer.time_step = time_step
    # save
    output_writer.save_results(net, time_step, pf_converged=pf_converged, ctrl_converged=ctrl_converged,
                               recycle_options=recycle_options)


def _call_output_writer(net, time_step, pf_converged, ctrl_converged, ts_variables):
    output_writer_routine(net, time_step, pf_converged, ctrl_converged, ts_variables['recycle_options'])


def run_time_step(net, time_step, ts_variables, run_control_fct=run_control, output_writer_fct=_call_output_writer,
                  **kwargs):
    """
    Time Series step function
    Is called to run the PANDAPOWER AC power flows with the timeseries module

    INPUT:
        **net** - The pandapower format network

        **time_step** (int) - time_step to be calculated

        **ts_variables** (dict) - contains settings for controller and time series simulation. See init_time_series()
    """
    ctrl_converged = True
    pf_converged = True
    # run time step function for each controller

    control_time_step(ts_variables['controller_order'], time_step)

    try:
        # calls controller init, control steps and run function (runpp usually is called in here)
        run_control_fct(net, ctrl_variables=ts_variables, **kwargs)
    except ControllerNotConverged:
        ctrl_converged = False
        # If controller did not converge do some stuff
        controller_not_converged(time_step, ts_variables)
    except ts_variables['errors']:
        # If power flow did not converge simulation aborts or continues if continue_on_divergence is True
        pf_converged = False
        pf_not_converged(time_step, ts_variables)

    output_writer_fct(net, time_step, pf_converged, ctrl_converged, ts_variables)

    finalize_step(ts_variables['controller_order'], time_step)


def _check_controller_recyclability(net):
    # if a parameter is set to True here, it will be recalculated during the time series simulation
    recycle = dict(trafo=False, gen=False, bus_pq=False)
    if "controller" not in net:
        # everything can be recycled since no controller is in net. But the time series simulation makes no sense
        # then anyway...
        return recycle

    for idx in net.controller.index:
        # todo: write to controller data frame recycle column instead of using self.recycle of controller instance
        ctrl_recycle = net.controller.at[idx, "recycle"]
        if not isinstance(ctrl_recycle, dict):
            # if one controller has a wrong recycle configuration it is deactived
            recycle = False
            break
        # else check which recycle parameter are set to True
        for rp in ["trafo", "bus_pq", "gen"]:
            recycle[rp] = recycle[rp] or ctrl_recycle[rp]

    return recycle


def _check_output_writer_recyclability(net, recycle, run):
    if "output_writer" not in net:
        raise ValueError("OutputWriter not defined")
    ow = net.output_writer.at[0, "object"]
    # results which are read with a faster batch function after the time series simulation
    recycle["batch_read"] = list()
    recycle["only_v_results"] = False
    new_log_variables = list()

    if hasattr(run, "__name__") and run.__name__ == "rundcpp":
        recycle["only_v_results"] = False
        recycle["batch_read"] = False
        return recycle

    for output in ow.log_variables:
        table, variable = output[0], output[1]
        if table not in ["res_bus", "res_line", "res_trafo", "res_trafo3w"] or recycle["trafo"] or len(output) > 2:
            # no fast read of outputs possible if other elements are required as these or tap changer is active
            recycle["only_v_results"] = False
            recycle["batch_read"] = False
            return recycle
        else:
            # fast read is possible
            if variable in ["vm_pu", "va_degree"]:
                new_log_variables.append(('ppc_bus', 'vm'))
                new_log_variables.append(('ppc_bus', 'va'))

            recycle["only_v_results"] = True
            recycle["batch_read"].append((table, variable))

    ow.log_variables = new_log_variables
    ow.log_variable('ppc_bus', 'vm')
    ow.log_variable('ppc_bus', 'va')
    return recycle


def get_recycle_settings(net, **kwargs):
    """
    checks if "run" is specified in kwargs and calls this function in time series loop.
    if "recycle" is in kwargs we use the TimeSeriesRunpp class (not implemented yet)

    INPUT:
        **net** - The pandapower format network

    RETURN:
        **recycle** - a dict with recycle options to be used by runpp
    """

    recycle = kwargs.get("recycle", None)
    if recycle is not False:
        # check if every controller can be recycled and what can be recycled
        recycle = _check_controller_recyclability(net)
        # if still recycle is not None, also check for fast output_writer features
        if recycle is not False:
            recycle = _check_output_writer_recyclability(net, recycle, kwargs.get("run", kwargs.get("run_control_fct")))

    return recycle


def init_time_steps(net, time_steps, **kwargs):
    # initializes time steps if as a range
    if not isinstance(time_steps, Iterable):
        if isinstance(time_steps, tuple):
            time_steps = range(time_steps[0], time_steps[1])
        elif time_steps is None and ("start_step" in kwargs and "stop_step" in kwargs):
            logger.warning("start_step and stop_step are depricated. "
                           "Please use a tuple like time_steps = (start_step, stop_step) instead or a list")
            time_steps = range(kwargs["start_step"], kwargs["stop_step"] + 1)
        else:
            logger.warning("No time steps to calculate are specified. "
                           "I'll check the datasource of the first controller for avaiable time steps")
            ds = net.controller.object.at[0].data_source
            if ds is None:
                raise UserWarning("No time steps are specified and the first controller doesn't have a data source"
                                  "the time steps could be retrieved from")
            else:
                max_timestep = ds.get_time_steps_len()
            time_steps = range(max_timestep)
    return time_steps


def init_time_series(net, time_steps, continue_on_divergence=False, verbose=True,
                     **kwargs):
    """
    inits the time series calculation
    creates the dict ts_variables, which includes necessary variables for the time series / control function

    INPUT:
        **net** - The pandapower format network

        **time_steps** (list or tuple, None) - time_steps to calculate as list or tuple (start, stop)
        if None, all time steps from provided data source are simulated

    OPTIONAL:

        **continue_on_divergence** (bool, False) - If True time series calculation continues in case of errors.

        **verbose** (bool, True) - prints progress bar or logger debug messages
    """

    time_steps = init_time_steps(net, time_steps, **kwargs)

    init_default_outputwriter(net, time_steps, **kwargs)
    # get run function
    run = kwargs.pop("run", runpp)
    recycle_options = None
    if hasattr(run, "__name__") and (run.__name__ == "runpp" or run.__name__ == "rundcpp"):
        # use faster runpp options if possible
        recycle_options = get_recycle_settings(net, run=run, **kwargs)

    init_output_writer(net, time_steps)
    # as base take everything considered when preparing run_control
    ts_variables = prepare_run_ctrl(net, None, run=run, **kwargs)
    # recycle options, which define what can be recycled
    ts_variables["recycle_options"] = recycle_options
    # time steps to be calculated (list or range)
    ts_variables["time_steps"] = time_steps
    # If True, a diverged run is ignored and the next step is calculated
    ts_variables["continue_on_divergence"] = continue_on_divergence
    # print settings
    ts_variables["verbose"] = verbose

    if logger.level != 10 and verbose:
        # simple progress bar
        ts_variables['progress_bar'] = tqdm.tqdm(total=len(time_steps))

    return ts_variables


def cleanup(net, ts_variables):
    if isinstance(ts_variables["recycle_options"], dict):
        # Todo: delete internal variables and dumped results which are not needed
        net._ppc = None  # remove _ppc because if recycle == True and a new timeseries calculation is started with a
        # different setup (in_service of lines or trafos, open switches etc.) it can lead to a disaster


def print_progress(i, time_step, time_steps, verbose, **kwargs):
    # simple status print in each time step.
    if logger.level != 10 and verbose:
        kwargs['ts_variables']["progress_bar"].update(1)

    # print debug info
    if logger.level == pplog.DEBUG and verbose:
        logger.debug("run time step %i" % time_step)

    # call a custom progress function
    if "progress_function" in kwargs:
        func = kwargs["progress_function"]
        func(i, time_step, time_steps, **kwargs)


def run_loop(net, ts_variables, run_control_fct=run_control, output_writer_fct=_call_output_writer, **kwargs):
    """
    runs the time series loop which calls runpp (or another run function) in each iteration

    Parameters
    ----------
    net - pandapower net
    ts_variables - settings for time series

    """
    for i, time_step in enumerate(ts_variables["time_steps"]):
        print_progress(i, time_step, ts_variables["time_steps"], ts_variables["verbose"], ts_variables=ts_variables,
                       **kwargs)
        run_time_step(net, time_step, ts_variables, run_control_fct, output_writer_fct, **kwargs)


def run_timeseries(net, time_steps=None, continue_on_divergence=False, verbose=True, check_controllers=True, **kwargs):
    """
    Time Series main function

    Runs multiple PANDAPOWER AC power flows based on time series which are stored in a **DataSource** inside
    **Controllers**. Optionally other functions than the pp power flow can be called by setting the run function in
    kwargs

    INPUT:
        **net** - The pandapower format network

    OPTIONAL:
        **time_steps** (list or tuple, None) - time_steps to calculate as list or tuple (start, stop)
        if None, all time steps from provided data source are simulated

        **continue_on_divergence** (bool, False) - If True time series calculation continues in case of errors.

        **verbose** (bool, True) - prints progress bar or if logger.level == Debug it prints debug messages

        **kwargs** - Keyword arguments for run_control and run If "run" is in kwargs the default call to runpp()
        is replaced by the function kwargs["run"]
    """

    ts_variables = init_time_series(net, time_steps, continue_on_divergence, verbose, **kwargs)

    # cleanup ppc before first time step
    cleanup(net, ts_variables)

    if check_controllers:
        control_diagnostic(net)  # produces significant overhead if you run many timeseries of short duration
    run_loop(net, ts_variables, **kwargs)

    # cleanup functions after the last time step was calculated
    cleanup(net, ts_variables)
    # both cleanups, at the start AND at the end, are important!



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import os

import numpy as np
import pandas as pd
import scipy.io

from pandapower.converter.pypower import from_ppc

try:
    from matpowercaseframes import CaseFrames

    matpowercaseframes_imported = True
except ImportError:
    matpowercaseframes_imported = False

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


def from_mpc(mpc_file, f_hz=50, casename_mpc_file='mpc', validate_conversion=False, **kwargs):
    """
    This function converts a matpower case file version 2 to a pandapower net.

    Note: If 'mpc_file' ends with '.m' the python package 'matpowercaseframes' is used. If
    'mpc_file' ends with '.mat' 'scipy.io.loadmat' is used. Other file endings are not supported.
    In that other cases, please, rename the file ending or use the internal subfunctions.

    Note: python is 0-based while Matlab is 1-based.

    INPUT:

        **mpc_file** - path to a matpower case file (.mat format not .m script).

    OPTIONAL:

        **f_hz** (int, 50) - The frequency of the network.

        **casename_mpc_file** (str, 'mpc') - The name of the variable in .mat file which contain
        the matpower case structure, i.e. the arrays "gen", "branch" and "bus".

        ****kwargs** - key word arguments for from_ppc()

    OUTPUT:

        **net** - The pandapower network

    EXAMPLE:
        >>> from pandapower.converter import from_mpc
        >>>
        >>> pp_net1 = from_mpc('case9.mat', f_hz=60)
        >>> pp_net2 = from_mpc('case9.m', f_hz=60)
    """
    ending = os.path.splitext(os.path.basename(mpc_file))[1]
    if ending == ".mat":
        ppc = _mat2ppc(mpc_file, casename_mpc_file)
    elif ending == ".m":
        ppc = _m2ppc(mpc_file, casename_mpc_file)
    net = from_ppc(ppc, f_hz=f_hz, validate_conversion=validate_conversion, **kwargs)
    if "mpc_additional_data" in ppc:
        if "_options" not in net:
            net["_options"] = dict()
        net._options.update(ppc["mpc_additional_data"])
        logger.info('added fields %s in net._options' % list(ppc["mpc_additional_data"].keys()))

    return net


def _mpc2ppc(mpc_file, casename_mpc_file):
    raise DeprecationWarning("_mpc2ppc() has been renamed by _mat2ppc().")


def _mat2ppc(mpc_file, casename_mpc_file):
    # load mpc from file
    mpc = scipy.io.loadmat(mpc_file, squeeze_me=True, struct_as_record=False)

    # init empty ppc
    ppc = dict()

    _copy_data_from_mpc_to_ppc(ppc, mpc, casename_mpc_file)
    _adjust_ppc_indices(ppc)
    _change_ppc_TAP_value(ppc)

    return ppc


def _m2ppc(mpc_file, casename_mpc_file):
    if not matpowercaseframes_imported:
        raise NotImplementedError(
            "matpowercaseframes is used to convert .m file. Please install that python "
            "package, e.g. via 'pip install matpowercaseframes'.")
    mpc_frames = CaseFrames(mpc_file)
    ppc = {key: mpc_frames.__getattribute__(key) if not isinstance(
        mpc_frames.__getattribute__(key), pd.DataFrame) else mpc_frames.__getattribute__(
        key).values for key in mpc_frames._attributes}
    _adjust_ppc_indices(ppc)
    return ppc


def _adjust_ppc_indices(ppc):
    # adjust indices of ppc, since ppc must start at 0 rather than 1 (matlab)
    ppc["bus"][:, 0] -= 1
    ppc["branch"][:, 0] -= 1
    ppc["branch"][:, 1] -= 1
    # if in ppc is only one gen -> numpy initially uses one dim array -> change to two dim array
    if len(ppc["gen"].shape) == 1:
        ppc["gen"] = np.array(ppc["gen"], ndmin=2)
    ppc["gen"][:, 0] -= 1


def _copy_data_from_mpc_to_ppc(ppc, mpc, casename_mpc_file):
    if casename_mpc_file in mpc:
        # if struct contains a field named mpc
        ppc['version'] = mpc[casename_mpc_file].version
        ppc["baseMVA"] = mpc[casename_mpc_file].baseMVA
        ppc["bus"] = mpc[casename_mpc_file].bus
        ppc["gen"] = mpc[casename_mpc_file].gen
        ppc["branch"] = mpc[casename_mpc_file].branch

        try:
            ppc['gencost'] = mpc[casename_mpc_file].gencost
        except:
            logger.info('gencost is not in mpc')

        for k in mpc[casename_mpc_file]._fieldnames:
            if k not in ppc:
                ppc.setdefault("mpc_additional_data", dict())[k] = getattr(mpc[casename_mpc_file], k)

    else:
        logger.error('Matfile does not contain a valid mpc structure.')


def _change_ppc_TAP_value(ppc):
    # adjust for the matpower converter -> taps should be 0 when there is no transformer, but are 1
    ppc["branch"][np.where(ppc["branch"][:, 8] == 0), 8] = 1


if "__main__" == __name__:
    pass



import inspect

import numpy as np
from numpy import complex128, zeros

from pandapower.auxiliary import _clean_up
from pandapower.build_branch import _calc_trafo_parameter, _calc_trafo3w_parameter
from pandapower.build_bus import _calc_pq_elements_and_add_on_ppc
from pandapower.control.controller.const_control import ConstControl
from pandapower.control.controller.trafo_control import TrafoController
from pandapower.pd2ppc import _pd2ppc
from pandapower.pf.pfsoln_numba import pfsoln as pfsoln_full, pf_solution_single_slack
from pandapower.pf.run_newton_raphson_pf import newtonpf, _get_pf_variables_from_ppci, _get_Y_bus, makeYbus_numba
from pandapower.powerflow import LoadflowNotConverged, _add_auxiliary_elements
from pandapower.pypower.idx_bus import PD, QD
from pandapower.pypower.idx_bus_dc import DC_PD
from pandapower.pypower.makeSbus import makeSbus
from pandapower.results import _copy_results_ppci_to_ppc, _extract_results, _get_aranged_lookup
from pandapower.results_branch import _get_branch_flows, _get_line_results, _get_trafo3w_results, _get_trafo_results
from pandapower.results_bus import _get_bus_results, _get_bus_dc_results
from pandapower.results_gen import _get_gen_results
from pandapower.run import runpp

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


class TimeSeriesRunpp:
    """
    Class for time series runpp.
    The goal of this class is a runpp function, which reuses many NR variables in each time step of a time series
    calculation.
    Also, not all power flow results are calculated in each runpp. Instead, only the results the output writer wants
    are calculated. Therefore, this class must be initialized with an output writer
    """

    def __init__(self, net):
        self.net = net
        self.output_writer = net.output_writer.iat[0, 0]

        # update functions
        self.update_pq = False
        self.update_trafo = False
        self.init_timeseries_newton()

    def ts_newtonpf(self, net):
        options = net["_options"]

        bus = self.ppci["bus"]
        branch = self.ppci["branch"]
        gen = self.ppci["gen"]
        svc = self.ppci["svc"]
        tcsc = self.ppci["tcsc"]
        ssc = self.ppci["ssc"]
        vsc = self.ppci["vsc"]
        # compute complex bus power injections [generation - load]
        # self.Cg = _get_Cg(gen_on, bus)
        # Sbus = _get_Sbus(self.baseMVA, bus, gen, self.Cg)
        Sbus = makeSbus(self.baseMVA, bus, gen)

        # run the newton power flow
        V, success, _, _, _, _, _ = newtonpf(Ybus=self.Ybus,
                                             Sbus=Sbus,
                                             V0=self.V,
                                             ref=self.ref,
                                             pv=self.pv,
                                             pq=self.pq,
                                             ppci=self.ppci,
                                             options=options)

        if not success:
            logger.warning("Loadflow not converged")
            logger.info("Lines of of service:")
            logger.info(net.line[~net.line.in_service])
            raise LoadflowNotConverged("Power Flow did not converge after")

        if self.ppci["gen"].shape[0] == 1 and not options["voltage_depend_loads"]:
            pfsoln = pf_solution_single_slack
        else:
            pfsoln = pfsoln_full

        bus, gen, branch = pfsoln(self.baseMVA, bus, gen, branch, svc, tcsc, ssc, vsc, self.Ybus, self.Yf, self.Yt, V,
                                  self.ref, self.ref_gens, Ibus=self.Ibus)

        self.ppci["bus"] = bus
        self.ppci["branch"] = branch
        self.ppci["gen"] = gen
        self.ppci["success"] = success
        self.ppci["et"] = None

        # ppci doesn't contain out of service elements, but ppc does -> copy results accordingly
        self.ppc = _copy_results_ppci_to_ppc(self.ppci, self.ppc, options["mode"])

        # raise if PF was not successful. If DC -> success is always 1
        if self.ppc["success"] != 1:
            _clean_up(net, res=False)
        else:
            net["_ppc"] = self.ppc
            net["converged"] = True

        self.V = V

        _extract_results(net, self.ppc)

        return net

    def _get_bus_p_q_results_from_ppc(self, net, ppc, net_bus_idx, ppc_bus_idx):
        """
        reads p, q results from ppc to net. Note: This function returns wrong vales if shunts, xwards and wards are in net
        @param net:
        @param ppc:
        @return:
        """
        # read bus_pq array which contains p and q values for each bus in net
        bus_pq = np.zeros(shape=(len(net["bus"].index), 2), dtype=float)
        bus_pq[net_bus_idx, 0] = ppc["bus"][ppc_bus_idx, PD] * 1e3
        bus_pq[net_bus_idx, 1] = ppc["bus"][ppc_bus_idx, QD] * 1e3
        bus_p_dc = np.zeros(shape=(len(net["bus_dc"].index), 1), dtype=np.float64)
        bus_dc_lookup_aranged = _get_aranged_lookup(net, "bus_dc")
        bus_p_dc[bus_dc_lookup_aranged, 0] = ppc["bus_dc"][:, DC_PD]  # todo test this

        bus_lookup_aranged = _get_aranged_lookup(net)
        _get_gen_results(net, ppc, bus_lookup_aranged, bus_pq)
        _get_bus_results(net, ppc, bus_pq)
        _get_bus_dc_results(net, bus_p_dc)

        net["res_bus"].index = net["bus"].index

    def get_branch_results(self, net, ppc, branch_update):
        i_ft, s_ft = _get_branch_flows(ppc)
        if "line" in branch_update:
            _get_line_results(net, ppc, i_ft)
        if "trafo" in branch_update:
            _get_trafo_results(net, ppc, s_ft, i_ft)
        if "trafo3w" in branch_update:
            _get_trafo3w_results(net, ppc, s_ft, i_ft)

    def init_timeseries_newton(self):
        """
        This function is called in the first iteration and variables needed in every loop are stored (Ybus, ppci...)
        @param net:
        @return:
        """
        self.init_newton_variables()
        net = self.net
        # get ppc and ppci
        # runpp(net, init_vm_pu="flat", init_va_degree="dc")
        # runpp(net, init_vm_pu="results", init_va_degree="results")
        runpp(net, init="dc")
        runpp(net, init="results")
        net._options["init_results"] = True
        net._options["init_vm_pu"] = "results"
        net._options["init_va_degree"] = "results"
        options = net._options
        _add_auxiliary_elements(net)
        self.ppc, self.ppci = _pd2ppc(net)
        net["_ppc"] = self.ppc

        self.baseMVA, bus, gen, branch, self.ref, self.pv, self.pq, _, _, self.V, self.ref_gens = \
            _get_pf_variables_from_ppci(self.ppci)
        self.ppci, self.Ybus, self.Yf, self.Yt = \
            _get_Y_bus(self.ppci, options, makeYbus_numba, self.baseMVA, bus, branch)
        self.Ibus = zeros(len(self.V), dtype=complex128)

        # self.Cg = _get_Cg(gen, bus)  # assumes that all gens are on!

        if "controller" in net:
            self.get_update_ctrl()

        return net

    def _update_nr_variables(self):
        """
        This function updates the stored NR variables for the powerflow, depending on the defined controllers for the
        time series calculation

        example:
        If only PQ updates are needed -> only pq values (in ppci.bus[:,PD]...) are updated
        If taps of trafos are updated -> Ybus must be recalculated
        Todo implement stuff for the controllers you use...

        @return:
        """
        ### PQ Updates
        if self.update_pq:
            # update P, Q values
            _calc_pq_elements_and_add_on_ppc(self.net, self.ppci)
            # adds P and Q for shunts, wards and xwards (to PQ nodes)
            # Todo: update shunts
            # _calc_shunts_and_add_on_ppc(self.net, self.ppci)

        ### Ybus recalculation
        if self.update_trafo:
            # update branch SHIFT entries for transformers (if tap changed)
            self.update_trafos()
            if "ybus_handler" in self.net:
                logger.error("Ybus update by trafo tap controller and ybus_handler is not possible simultaneously")

        if "ybus_handler" in self.net:
            self.Ybus, self.Yf, self.Yt = self.net["ybus_handler"].get_y()

    def cleanup(self):
        _clean_up(self.net)
        self.init_newton_variables()

    def ts_runpp(self, net, **kwargs):
        # update pq values in ppci
        self._update_nr_variables()

        # ----- run the powerflow -----
        net = self.ts_newtonpf(net)

        return net

    def init_newton_variables(self):
        self.baseMVA = None
        self.ref = None
        self.pv = None
        self.pq = None
        self.V = None
        self.ppc = None
        self.ppci = None
        self.Ybus = None
        self.Yf = None
        self.Yt = None
        self.Ibus = None
        self.Cg = None

    def update_trafos(self):
        net = self.net
        ppci = self.ppci

        # update branch SHIFT entries for transfomers (if tap changed)
        lookup = net._pd2ppc_lookups["branch"]
        if "trafo" in lookup:
            _calc_trafo_parameter(net, ppci)
        if "trafo3w" in lookup:
            _calc_trafo3w_parameter(net, ppci)

        # update Ybus based on this
        options = net._options
        baseMVA, bus, gen, branch, svc, tcsc, ssc, vsc, ref, pv, pq, _, _, V, _ = _get_pf_variables_from_ppci(ppci)
        self.ppci, self.Ybus, self.Yf, self.Yt = _get_Y_bus(ppci, options, makeYbus_numba, baseMVA, bus,
                                                            branch)

    def get_update_ctrl(self):
        controllers = self.net.controller["object"]
        update_pq = False
        update_trafo = False
        for controller in controllers:
            base_clases = inspect.getmro(controller.__class__)
            if ConstControl in base_clases:
                update_pq = True
            elif TrafoController in controller.__class__.__bases__:
                update_trafo = True
            else:
                raise TypeError("controller class not supported for recycle %s" % controller.__class__)

        self.update_pq = update_pq
        self.update_trafo = update_trafo



from logging import StreamHandler, Formatter

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging


class AppHandler(StreamHandler):
    def __init__(self, app, freeze_app_between_messages=False):
        super().__init__()
        self.app = app
        self.PrintPlain = app.PrintPlain
        self.PrintInfo = app.PrintInfo
        self.PrintWarn = app.PrintWarn
        self.PrintError = app.PrintError
        self.name = 'PowerFactory Converter'
        formatter = Formatter('%(message)s')
        self.setFormatter(formatter)
        self.freeze_app_between_messages = freeze_app_between_messages

    def emit(self, record):
        if self.freeze_app_between_messages:
            self.app.SetGuiUpdateEnabled(1)

        msg = self.format(record)
        level = record.levelname
        if level == "DEBUG":
            self.PrintPlain(msg)
        elif level == "INFO":
            self.PrintInfo(msg)
        elif level == "WARNING":
            self.PrintWarn(msg)
        elif level == "ERROR":
            self.PrintError(msg)
        elif level == 'CRITICAL':
            self.PrintError(msg)
        else:
            self.PrintPlain(msg)

        if self.freeze_app_between_messages:
            self.app.SetGuiUpdateEnabled(0)


def setup_logger(app, name, level):
    logger = logging.getLogger(name)

    app_handler = AppHandler(app)
    logger.addHandler(app_handler)

    set_PF_level(logger, app_handler, level)

    logger.info('initialized logger %s' % logger.name)
    return logger, app_handler


def set_PF_level(logger, app_handler, level):
    if level == 'DEBUG':
        logger.setLevel(logging.DEBUG)
        app_handler.setLevel(logging.DEBUG)
    elif level == 'INFO':
        logger.setLevel(logging.INFO)
        app_handler.setLevel(logging.INFO)
    elif level == 'WARNING':
        logger.setLevel(logging.WARNING)
        app_handler.setLevel(logging.WARNING)
    elif level == 'ERROR':
        logger.setLevel(logging.ERROR)
        app_handler.setLevel(logging.ERROR)



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

from numpy import roots, conj, r_, exp


def _iwamoto_step(Ybus, J, F, dx, pq, npv, npq, dVa, dVm, Vm, Va, pv, j1, j2, j3, j4, j5, j6):
    if npv:
        dVa[pv] = dx[j1:j2]
    if npq:
        dVa[pq] = dx[j3:j4]
        dVm[pq] = dx[j5:j6]
    dV = dVm * exp(1j * dVa)

    iwa_multiplier = _get_iwamoto_multiplier(Ybus, J, F, dV, dx, pq, pv)

    Vm += iwa_multiplier * dVm
    Va += iwa_multiplier * dVa
    return Vm, Va


def _get_iwamoto_multiplier(Ybus, J, F, dV, dx, pq, pv):
    """
    Calculates the iwamato multiplier to increase convergence
    """

    c0=-F                               # c0 = ys-y(x)= -F
    c1=-J * dx                          # c1 = -Jdx
    c2=-_evaluate_Yx(Ybus, dV, pv, pq)  # c2 = -y(dx)

    g0 = c0.dot(c1)
    g1 = c1.dot(c1) + 2 * c0.dot(c2)
    g2 = 3.0 * c1.dot(c2)
    g3 = 2.0 * c2.dot(c2)

    np_roots = roots([g3, g2, g1, g0])[2].real
    print("iwamoto muliplier:", np_roots)
    #print(g0,g1,g2,g3,np_roots)
    #print(g0+ g1*a + g2*a**2 + g3*a**3)

    return np_roots


def _evaluate_Yx(Ybus, V, pv, pq):
    ## evaluate y(x)
    Yx = V * conj(Ybus * V)
    F = r_[Yx[pv].real,
           Yx[pq].real,
           Yx[pq].imag]
    return F



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

from typing import List, Tuple, TYPE_CHECKING, Dict, Any, Union

import numpy as np

# TYPE_CHECKING is used to avoid circular imports, see https://stackoverflow.com/a/39757388
if TYPE_CHECKING:
    import pandapipes
from typing_extensions import deprecated

import sys
import math
import pandas as pd
from numpy import array

from pandapower.auxiliary import soft_dependency_error, pandapowerNet

# get logger (same as in simple_plot)
try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)

try:
    from shapely.geometry import Point, LineString

    shapely_INSTALLED = True
except ImportError:
    shapely_INSTALLED = False

try:
    from geopandas import GeoDataFrame, GeoSeries

    geopandas_INSTALLED = True
except ImportError:
    geopandas_INSTALLED = False

try:
    from pyproj import Transformer

    pyproj_INSTALLED = True
except ImportError:
    pyproj_INSTALLED = False

try:
    import geojson

    geojson_INSTALLED = True
except ImportError:
    geojson_INSTALLED = False


def _node_geometries_from_geodata(node_geo, epsg=31467, remove_xy=False):
    """
    Creates a geopandas geodataframe from a given dataframe of with node coordinates as x and y
    values.

    :param node_geo: The dataframe containing the node coordinates (x and y values)
    :type node_geo: pandas.dataframe
    :param epsg: The epsg projection of the node coordinates
    :type epsg: int, default 31467 (= Gauss-Krüger Zone 3)
    :param remove_xy: If x/y and coords columns should be removed from the geodataframe
    :type remove_xy: bool, default False
    :return: node_geodata - a geodataframe containing the node_geo and Points in the geometry column
    """
    missing_packages = array(["shapely", "geopandas"])[~array([
        shapely_INSTALLED, geopandas_INSTALLED])]
    if len(missing_packages):
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", missing_packages)
    geoms = [Point(x, y) for x, y in node_geo[["x", "y"]].values]
    if remove_xy:
        return GeoDataFrame(crs=f"epsg:{epsg}", geometry=geoms, index=node_geo.index)
    return GeoDataFrame(node_geo, crs=f"epsg:{epsg}", geometry=geoms, index=node_geo.index)


def _branch_geometries_from_geodata(branch_geo, epsg=31467, remove_xy=False):
    missing_packages = array(["shapely", "geopandas"])[~array([
        shapely_INSTALLED, geopandas_INSTALLED])]
    if len(missing_packages):
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", missing_packages)
    geoms = GeoSeries([LineString(x) for x in branch_geo.coords.values], index=branch_geo.index,
                      crs=f"epsg:{epsg}")
    if remove_xy:
        return GeoDataFrame(crs=f"epsg:{epsg}", geometry=geoms, index=branch_geo.index)
    return GeoDataFrame(branch_geo, crs=f"epsg:{epsg}", geometry=geoms, index=branch_geo.index)


def _transform_node_geometry_to_geodata(node_geo):
    """
    Create x and y values from geodataframe

    :param node_geo: The dataframe containing the node geometries (as shapely points)
    :type node_geo: geopandas.GeoDataFrame
    :return: bus_geo - The given geodataframe with x and y values
    """
    node_geo["x"] = [p.x for p in node_geo.geometry]
    node_geo["y"] = [p.y for p in node_geo.geometry]
    return node_geo


def _transform_branch_geometry_to_coords(branch_geo):
    """
    Create coords entries from geodataframe geometries

    :param branch_geo: The dataframe containing the branch geometries (as shapely LineStrings)
    :type branch_geo: geopandas.GeoDataFrame
    :return: branch_geo - The given geodataframe with coords
    """
    branch_geo["coords"] = branch_geo.geometry.apply(lambda x: list(x.coords))
    return branch_geo


def _transform_node_geometry_to_geojson(node_geo):
    """
    Create geojson string from geodataframe geometries

    ! forces projection to wgs84 !

    :param node_geo: The dataframe containing the node geometries (as shapely points)
    :type node_geo: geopandas.GeoDataFrame
    :return: A geojson string for the node_geo
    """
    return node_geo.to_json(na='drop', show_bbox=True, drop_id=False, to_wgs84=True)


def _transform_branch_geometry_to_geojson(branch_geo):
    """
    Create geojson from geodataframe geometries

    ! forces projection to wgs84 !

    :param branch_geo: The dataframe containing the branch geometries (as shapely LineStrings)
    :type branch_geo: geopandas.GeoDataFrame
    :return: A geojson object for the branch_geo
    """
    return branch_geo.to_json(na='drop', show_bbox=True, drop_id=False, to_wgs84=True)


def _convert_xy_epsg(x, y, epsg_in=4326, epsg_out=31467):
    """
    Converts the given x and y coordinates according to the defined epsg projections.

    :param x: x-values of coordinates
    :type x: iterable
    :param y: y-values of coordinates
    :type y: iterable
    :param epsg_in: current epsg projection
    :type epsg_in: int, default 4326 (= WGS84)
    :param epsg_out: epsg projection to be transformed to
    :type epsg_out: int, default 31467 (= Gauss-Krüger Zone 3)
    :return: transformed_coords - x and y values in new coordinate system
    """
    if not pyproj_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "pyproj")
    transformer = Transformer.from_crs(f'EPSG:{epsg_in}', f'EPSG:{epsg_out}', always_xy=True)
    return transformer.transform(x, y)


@deprecated("Use convert_gis_to_geojson instead. Support for geodata will be dropped.")
def convert_gis_to_geodata(net, node_geodata=True, branch_geodata=True):
    """
    Extracts information on bus and line geodata from the geometries of a geopandas geodataframe.

    :param net: The net for which to convert the geodata
    :type net: pandapowerNet
    :param node_geodata: flag if to extract x and y values for bus geodata
    :type node_geodata: bool, default True
    :param branch_geodata: flag if to extract coordinates values for line geodata
    :type branch_geodata: bool, default True
    :return: No output.
    """
    if node_geodata:
        _transform_node_geometry_to_geodata(net.bus_geodata)
    if branch_geodata:
        _transform_branch_geometry_to_coords(net.line_geodata)


@deprecated(
    "Use convert_geodata_to_geojson instead. Support for gis will be dropped.\
    To get a geodataframe use GeoDataFrame.from_features."
)
def convert_geodata_to_gis(net, epsg=31467, node_geodata=True, branch_geodata=True, remove_xy=False):
    """
    Transforms the bus and line geodata of a net into a geopandas geodataframe with the respective
    geometries.

    :param net: The net for which to convert the geodata
    :type net: pandapowerNet
    :param epsg: current epsg projection
    :type epsg: int, default 4326 (= WGS84)
    :param node_geodata: flag if to transform the bus geodata table
    :type node_geodata: bool, default True
    :param branch_geodata: flag if to transform the line geodata table
    :type branch_geodata: bool, default True
    :param remove_xy: flag if to remove x,y and coords columns from geodata tables
    :return: No output.
    """
    converted = False
    if node_geodata and "bus_geodata" in net:
        net["bus_geodata"] = _node_geometries_from_geodata(net["bus_geodata"], epsg, remove_xy)
        converted = True
    if branch_geodata and "line_geodata" in net:
        net["line_geodata"] = _branch_geometries_from_geodata(net["line_geodata"], epsg, remove_xy)
        converted = True
    if converted:
        net["gis_epsg_code"] = epsg


@deprecated("Use convert_crs instead. Networks should not use different crs for bus and line geodata.")
def convert_epsg_bus_geodata(net, epsg_in=4326, epsg_out=31467):
    """
    Converts bus geodata in net from epsg_in to epsg_out

    :param net: The pandapower network
    :type net: pandapowerNet
    :param epsg_in: current epsg projection
    :type epsg_in: int, default 4326 (= WGS84)
    :param epsg_out: epsg projection to be transformed to
    :type epsg_out: int, default 31467 (= Gauss-Krüger Zone 3)
    :return: net - the given pandapower network (no copy!)
    """
    net['bus_geodata'].loc[:, "x"], net['bus_geodata'].loc[:, "y"] = _convert_xy_epsg(
        net['bus_geodata'].loc[:, "x"], net['bus_geodata'].loc[:, "y"], epsg_in, epsg_out)
    return net


def convert_crs(net: pandapowerNet or 'pandapipes.pandapipesNet', epsg_in=4326, epsg_out=31467):
    """
    This function works for pandapowerNet and pandapipesNet. Documentation will refer to names from pandapower.
    Converts bus and line geodata in net from epsg_in to epsg_out
    if GeoDataFrame data is present convert_geodata_to_gis should be used to update geometries after crs conversion

    :param net: The pandapower network
    :type net: pandapowerNet|pandapipesNet
    :param epsg_in: current epsg projection
    :type epsg_in: int, default 4326 (= WGS84)
    :param epsg_out: epsg projection to be transformed to
    :type epsg_out: int, default 31467 (= Gauss-Krüger Zone 3)
    :return: net - the given pandapower network (no copy!)
    """
    is_pandapower = net.__class__.__name__ == 'pandapowerNet'
    if epsg_in == epsg_out:
        return

    if ('geo' in net.bus and not all(net.bus.geo.isna()) and
            'geo' in net.line and not all(net.line.geo.isna()) and
            epsg_out == 4326):
        # by definition geojson is in wgs84
        return

    if not pyproj_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "pyproj")
    transformer = Transformer.from_crs(epsg_in, epsg_out, always_xy=True)

    def _geo_node_transformer(r):
        (x, y) = transformer.transform(r.x, r.y)
        if is_pandapower:
            coords = r.coords
            if coords and not pd.isna(coords):
                coords = _geo_branch_transformer(coords)
            return pd.Series([x, y, coords], ["x", "y", "coords"])
        else:
            return pd.Series([x, y], ["x", "y"])

    def _geo_branch_transformer(r):
        return list(transformer.itransform(r))

    if is_pandapower:
        net.bus_geodata = net.bus_geodata.apply(lambda r: _geo_node_transformer(r), axis=1)
        net.line_geodata.coords = net.line_geodata.coords.apply(lambda r: _geo_branch_transformer(r))
        net.bus_geodata.attrs = {"crs": f"EPSG:{epsg_out}"}
        net.line_geodata.attrs = {"crs": f"EPSG:{epsg_out}"}
    else:
        net.junction_geodata = net.junction_geodata.apply(lambda r: _geo_node_transformer(r), axis=1)
        net.pipe_geodata.coords = net.pipe_geodata.coords.apply(lambda r: _geo_branch_transformer(r))
        net.junction_geodata.attrs = {"crs": f"EPSG:{epsg_out}"}
        net.pipe_geodata.attrs = {"crs": f"EPSG:{epsg_out}"}


def dump_to_geojson(
        net: pandapowerNet or 'pandapipes.pandapipesNet',
        nodes: Union[bool, List[int]] = False,
        branches: Union[bool, List[int]] = False,
        switches: Union[bool,  List[int]] = False,
        trafos: Union[bool, List[int]] = False,
        t_is_3w: bool = False,
        include_type_id: bool = True
) -> geojson.FeatureCollection:
    """
    This function works for pandapowerNet and pandapipesNet. Documentation will refer to names from pandapower.
    Dumps all primitive values from bus, bus_geodata, res_bus, line, line_geodata and res_line into a geojson object.
    It is recommended to only dump networks using WGS84 for GeoJSON specification compliance.

    Since switches and trafos do not contain their own geodata in pandapower, geodata is taken from the components
    connected to them. Trafos are always given the geodata from the lv_bus connected at the trafo. Switches are given
    geodata from the bus the switch is connected to. They do not carry info about where the other components connected
    to them are located!

    :param net: The pandapower network
    :type net: pandapowerNet|pandapipesNet
    :param nodes: if True return contains all bus data, can be a list of bus ids that should be contained
    :type nodes: bool | list, default False
    :param branches: if True return contains all line data, can be a list of line ids that should be contained
    :type branches: bool | list, default False
    :param switches: if True return contains all switch data, can be a list of switch ids that should be contained (only supported for pandapowerNet)
    :type switches: bool | list, default False
    :param trafos: if True return contains all trafo data, can be a list of trafo ids that should be contained (only supported for pandapowerNet)
    :type trafos: bool | list, default False
    :param t_is_3w: if True, the trafos are treated as 3W-trafos
    :type t_is_3w: bool, default False
    :param include_type_id: if True, pp_type and pp_index is added to every feature
    :type include_type_id: bool, default true
    :return: A geojson object.
    :return type: geojson.FeatureCollection
    """
    is_pandapower = net.__class__.__name__ == 'pandapowerNet'

    if not geojson_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "geojson")

    try:
        if is_pandapower:
            if hasattr(net, "bus_geodata") or hasattr(net, "line_geodata"):
                raise UserWarning("""The supplied network uses an outdated geodata format. Please update your geodata by
                                     \rrunning `pandapower.plotting.geo.convert_geodata_to_geojson(net)`""")
            else:
                node_geodata = net.bus.geo
                branch_geodata = net.line.geo
        else:
            if hasattr(net, "junction_geodata") or hasattr(net, "pipe_geodata"):
                raise UserWarning("""The supplied network uses an outdated geodata format. Please update your geodata by
                                     \rrunning `pandapower.plotting.geo.convert_geodata_to_geojson(net)`""")
            else:
                node_geodata = net.junction.geo
                branch_geodata = net.pipe.geo
    except UserWarning as e:
        logger.warning(e)
        return geojson.FeatureCollection([])

    def _get_props(r, c, p) -> None:
        for col in c:
            try:
                p[col] = float(r[col])
                if math.isnan(p[col]):
                    p[col] = None
            except (ValueError, TypeError):
                p[col] = str(r[col])

    def update_props(r: pd.Series) -> None:
        if r.name not in props:
            props[r.name] = {}
        props[r.name].update(r.to_dict())

    missing_geom: List[int] = [0, 0, 0, 0]  # missing nodes, branches, switches, trafos
    features = []
    # build geojson features for nodes
    if nodes:
        props = {}
        for table in (['bus', 'res_bus'] if is_pandapower else ['junction', 'res_junction']):
            if table not in net.keys():
                continue

            tempdf = net[table].copy(deep=True)
            if include_type_id:
                tempdf['pp_type'] = 'bus' if is_pandapower else 'junction'
                tempdf['pp_index'] = tempdf.index
            tempdf.index = tempdf.apply(lambda r: f"{r['pp_type']}-{r['pp_index']}", axis=1)
            tempdf.drop(columns=['geo'], inplace=True, axis=1, errors='ignore')

            tempdf.apply(update_props, axis=1)
        if isinstance(nodes, bool):
            iterator = node_geodata.items()
        else:
            iterator = node_geodata.loc[nodes].items()
        for ind, geom in iterator:
            if geom is None or pd.isna(geom) or geom == "[]":
                missing_geom[0] += 1
                continue
            uid = f"{'bus' if is_pandapower else 'junction'}-{ind}"
            features.append(geojson.Feature(geometry=geojson.loads(geom), id=uid, properties=props[uid]))

    # build geojson features for branches
    if branches:
        props = {}
        for table in (['line', 'res_line'] if is_pandapower else ['pipe', 'res_pipe']):
            if table not in net.keys():
                continue

            tempdf = net[table].copy(deep=True)
            if include_type_id:
                tempdf['pp_type'] = 'line' if is_pandapower else 'pipe'
                tempdf['pp_index'] = tempdf.index
            tempdf.index = tempdf.apply(lambda r: f"{r['pp_type']}-{r['pp_index']}", axis=1)
            tempdf.drop(columns=['geo'], inplace=True, axis=1, errors='ignore')

            tempdf.apply(update_props, axis=1)

        # Iterating over pipe_geodata won't work
        # pipe_geodata only contains pipes that have inflection points!
        if isinstance(branches, bool):
            # if all iterating over pipe
            iterator = branch_geodata.items()
        else:
            iterator = branch_geodata.loc[branches].items()
        for ind, geom in iterator:
            if geom is None or pd.isna(geom) or geom == "[]":
                missing_geom[1] += 1
                continue
            uid = f"{'line' if is_pandapower else 'pipe'}-{ind}"
            features.append(geojson.Feature(geometry=geojson.loads(geom), id=uid, properties=props[uid]))

    if switches and is_pandapower:
        if isinstance(switches, bool):
            switches = net.switch.index
        if 'switch' in net.keys():
            cols = net.switch.columns
            for ind, row in net.switch.loc[switches].iterrows():
                if pd.isna(row.bus):
                    # switch is not connected to a bus! Will count this as missing geometry.
                    missing_geom[2] += 1
                    continue
                prop = {}
                if include_type_id:
                    prop = {
                        'pp_type': 'switch',
                        'pp_index': ind,
                    }
                uid = f"switch-{ind}"
                _get_props(row, cols, prop)

                # getting geodata for switches
                geom = geojson.loads(net.bus.geo.at[row.bus])
                if isinstance(geom, geojson.LineString):
                    logger.warning(f"LineString geometry not supported for type 'switch'. Skipping switch {ind}")
                    geom = None
                if geom is None or geom == "[]":
                    missing_geom[2] += 1
                    continue
                features.append(geojson.Feature(geometry=geom, id=uid, properties=prop))

        if trafos and is_pandapower:
            t_type = 'trafo3w' if t_is_3w else 'trafo'
            if isinstance(trafos, bool):
                trafos = net[t_type].index
            if t_type in net.keys():
                cols = net[t_type].columns
                for ind, row in net[t_type].loc[trafos].iterrows():
                    prop = {}
                    if include_type_id:
                        prop = {
                            'pp_type': t_type,
                            'pp_index': ind,
                        }
                    uid = f"{t_type}-{ind}"
                    _get_props(row, cols, prop)

                    # getting geodata for switches
                    geom = geojson.loads(net.bus.geo.at[row.lv_bus])
                    if isinstance(geom, geojson.LineString):
                        logger.warning(f"LineString geometry not supported for type '{t_type}'. Skipping trafo {ind}")
                    if geom is None or geom == "[]":
                        missing_geom[3] += 1
                        continue
                    features.append(geojson.Feature(geometry=geom, id=uid, properties=prop))

    if any(missing_geom):
        missing_str = []
        if missing_geom[0]:
            missing_str.append(f"{missing_geom[0]} branch geometries")
        if missing_geom[1]:
            missing_str.append(f"{missing_geom[1]} node geometries")
        if missing_geom[2]:
            missing_str.append(f"{missing_geom[2]} switch geometries")
        if missing_geom[3]:
            missing_str.append(f"{missing_geom[3]} trafo geometries")
        logger.warning(f"{', '.join(missing_str)} could not be converted to geojson. Please update network's geodata!")

    # find and set crs if available
    crs_node = None
    if nodes and "crs" in node_geodata.attrs:
        crs_node = node_geodata.attrs["crs"]
    crs_branch = None
    if branches and "crs" in branch_geodata.attrs:
        crs_branch = branch_geodata.attrs["crs"]

    crs = {
        "type": "name",
        "properties": {
            "name": ""
        }
    }
    if crs_node:
        if crs_branch and crs_branch != crs_node:
            raise ValueError("Node and Branch crs mismatch")
        crs["properties"]["name"] = crs_node
    elif crs_branch:
        crs["properties"]["name"] = crs_branch
    else:
        crs = None
    if crs:
        return geojson.FeatureCollection(features, crs=crs)
    return geojson.FeatureCollection(features)


def convert_geodata_to_geojson(
        net: pandapowerNet or 'pandapipes.pandapipesNet',
        delete: bool = True,
        lonlat: bool = False) -> None:
    """
    Converts bus_geodata and line_geodata to bus.geo and line.geo column entries.
    If used on pandapipesNet, the junction_geodata and pipe_geodata are converted.

    It is expected that any input network has its coords in WGS84 (epsg:4326) projection.
    If this is not the case use convert_crs to convert the network to WGS84.

    :param net: The pandapower network containing line_geodata and bus_geodata in WGS84!
    :type net: pandapowerNet
    :param delete: If True, the geodataframes are deleted after conversion
    :type delete: bool, default True
    :param lonlat: If True, the coordinates are expected to be in lonlat format (x=lon, y=lat)
    :type lonlat: bool, default False
    """
    is_pandapower = net.__class__.__name__ == 'pandapowerNet'

    if is_pandapower:
        df = net.bus
        ldf = net.line
        geo_df = net.bus_geodata if (hasattr(net, 'bus_geodata') and isinstance(net.bus_geodata, pd.DataFrame)) else pd.DataFrame()
        geo_ldf = net.line_geodata if (hasattr(net, 'line_geodata') and isinstance(net.line_geodata, pd.DataFrame)) else pd.DataFrame()
    else:
        df = net.junction
        ldf = net.pipe
        geo_df = net.junction_geodata if (hasattr(net, 'junction_geodata') and isinstance(net.junction_geodata, pd.DataFrame)) else pd.DataFrame()
        geo_ldf = net.pipe_geodata if (hasattr(net, 'pipe_geodata') and isinstance(net.pipe_geodata, pd.DataFrame)) else pd.DataFrame()

    a, b = "yx" if lonlat else "xy"  # substitute x and y with a and b to reverse them if necessary
    if not geo_df.empty:
        df["geo"] = geo_df.apply(lambda r: f'{{"coordinates": [{r[a]}, {r[b]}], "type": "Point"}}', axis=1)

    ldf["geo"] = np.nan
    for l_id in ldf.index:
        if l_id not in geo_ldf.index:
            continue
        # pandapipes currently only stores inflection points for pipes. This function will inject start and end points.
        if is_pandapower:
            coords: List[List[float]] = [[y, x] if lonlat else [x, y] for x, y in geo_ldf.coords.at[l_id]]
        else:
            coords: List[List[float]] = []
            from_coords = geo_df.loc[ldf[l_id].from_junction]
            to_coords = geo_df.loc[ldf[l_id].to_junction]
            coords.append([float(from_coords.x), float(from_coords.y)])
            if l_id in net.pipe_geodata:
                coords.append(geo_ldf.loc[l_id].coords)
            coords.append([float(to_coords.x), float(to_coords.y)])
        if not coords:
            continue
        ls = f'{{"coordinates": {coords}, "type": "LineString"}}'
        ldf["geo"] = ldf["geo"].astype(object)
        ldf.geo.at[l_id] = ls

    if delete:
        if is_pandapower:
            if hasattr(net, 'bus_geodata'):del net.bus_geodata
            if hasattr(net, 'line_geodata'): del net.line_geodata
        else:
            if hasattr(net, 'junction_geodata'): del net.junction_geodata
            if hasattr(net, 'pipe_geodata'): del net.pipe_geodata


def convert_gis_to_geojson(
        net: pandapowerNet or 'pandapipes.pandapipesNet',
        delete: bool = True) -> None:
    """
    Transforms the bus and line geodataframes of a net into a geojson object.

    :param net: The net for which to convert the geodataframes
    :type net: pandapowerNet
    :param delete: If True, the geodataframes are deleted after conversion
    :type delete: bool, default True
    :return: No output.
    """

    is_pandapower = net.__class__.__name__ == 'pandapowerNet'

    if is_pandapower:
        net.bus["geo"] = _transform_node_geometry_to_geojson(net["bus_geodata"])
        net.line["geo"] = _transform_branch_geometry_to_geojson(net["line_geodata"])

        if delete:
            del net.bus_geodata
            del net.line_geodata
    else:
        net.junction["geo"] = _transform_node_geometry_to_geojson(net["junction_geodata"])
        net.pipe["geo"] = _transform_branch_geometry_to_geojson(net["pipe_geodata"])

        if delete:
            del net.junction_geodata
            del net.pipe_geodata



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

"""Defines constants for named column indices to SVC matrix.

Some examples of usage, after defining the constants using the lines above,
are::

    svc[3, SVC_STATUS] = 0              # take SVC 4 out of service
    SVC_Q = svc[:, SVC_Q]               # extract reactive power vector

The index, name, and meaning of each column of the SVC matrix are given below:

columns 0-10 must be included in the input matrix (in case file)
    0.  C{SVC_BUS}                          SVC bus number
    1.  C{SVC_X_L}                          inductive reactance (p.u.)
    2.  C{SVC_X_CVAR}                       capacitive reactance (p.u.)
    3.  C{SVC_SET_VM_PU}                    SVC bus voltage magnitude set point (p.u.)
    4.  C{SVC_THYRISTOR_FIRING_ANGLE}       SVC thyristor firing angle (radians)
    5.  C{SVC_STATUS}                       initial SVC status, 1 - in service, 0 - out of service
    6.  C{SVC_CONTROLLABLE}                 controllability flag for SVC
    7.  C{SVC_MIN_FIRING_ANGLE}             minimum thyristor firing angle (radians)
    8.  C{SVC_MAX_FIRING_ANGLE}             maximum thyristor firing angle (radians)
    9.  C{SVC_Q}                            reactive power absorbed (+) or injected (-) by SVC (MVAr)
    10. C{SVC_X_PU}                         reactance in per unit (p.u.)

"""


# define the indices
SVC_BUS = 0  # f, from bus number
SVC_X_L = 1  # (p.u.)
SVC_X_CVAR = 2  # (p.u.)
SVC_SET_VM_PU = 3  # (p.u.)
SVC_THYRISTOR_FIRING_ANGLE = 4
SVC_STATUS = 5  # initial branch status, 1 - in service, 0 - out of service
SVC_CONTROLLABLE = 6
SVC_MIN_FIRING_ANGLE = 7
SVC_MAX_FIRING_ANGLE = 8

SVC_Q = 9
SVC_X_PU = 10

svc_cols = 11
